import argparse
import sys
import os
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
import matplotlib.cm as cm
import umap
import pandas as pd
from scipy import stats
from scipy.stats.mstats import f_oneway as aov
from scipy.stats import ttest_ind as ttest
from scipy.stats import chisquare
from scipy.stats import chi2,chi2_contingency
from scipy.stats.contingency import margins
from copy import deepcopy

from mana_clust.common_functions import write_table, process_dir, get_sample_k_lists, import_dict, correct_pvalues_for_multiple_testing, do_louvain_merger
from mana_clust.mana_clust import *

##########################################################
## stats functions


def get_anova(expression_vect, group_vect, num_in_each_group):
    global full_expression, list_of_k_sample_indices
    if np.sum(full_expression[index,:]) == 0:
        return([np.nan,np.nan])
    list_of_group_values = []## keeps track of the values of each group's expression values
    for group in range(0,len(list_of_k_sample_indices)):
        if len(list_of_k_sample_indices[group]) > 0:
            list_of_group_values.append(full_expression[index,list_of_k_sample_indices[group]])
    ## go through and check for empty lists
    return_val = list(aov(*list_of_group_values))
    return(return_val)




###############################################################

def get_per_ome_feat_select_labels(subjects, meta_ome, ome_name = ''):
    feat_labels = []
    for subj in subjects:
        if subj in meta_ome.label_dict:
            feat_labels.append(meta_ome.label_dict[subj])
        else:
            sys.exit("there is a subject in ome:\n\t"+ome_name+"\nthat we couldn't find in the meta_ome")
    return(feat_labels)


def get_within_ome_similarity(feat_select_labels, ome, out_dir):
    print("\ttesting whether there are overall pairwise differences in affinities")
    #print(feat_select_labels)
    F_mat = np.zeros((len(feat_select_labels),len(feat_select_labels)))
    p_mat = np.zeros((len(feat_select_labels),len(feat_select_labels)))
    ## set the diagonal to one beacuse everything is equal to itself
    for i in range(p_mat.shape[0]):
        p_mat[i,i]=1.0
    for i in range(0,len(feat_select_labels)):
        first_indices = feat_select_labels[i]
        for j in range(i, len(feat_select_labels)):
            if i != j:
                second_indices = feat_select_labels[j]
                if len(second_indices)>=3 and len(first_indices)>=3:
                    ## subset within group affinities
                    #print(ome.affinity_matrix[:3,:3])
                    temp_first_within = ome.affinity_matrix[first_indices,:]
                    temp_first_within = temp_first_within[:,first_indices]
                    #print(temp_first_within[:3,:3])
                    temp_second_within = ome.affinity_matrix[second_indices,:]
                    temp_second_within = temp_second_within[:,second_indices]
                    ## to remove duplicates (from matrix symmetry), just get the upper triangle
                    temp_first_within = temp_first_within[np.tril_indices(temp_first_within.shape[0], k=-1)].tolist()
                    temp_second_within = temp_second_within[np.tril_indices(temp_second_within.shape[0], k=-1)].tolist()
                    ## combine them
                    combined_within_group_affinities = temp_first_within + temp_second_within
                    #print(combined_within_group_affinities)
                    
                    ## now get the across group affinities
                    across_group_affinities = ome.affinity_matrix[first_indices,:]
                    across_group_affinities = across_group_affinities[:,second_indices].flatten().tolist()

                    f, p = list(aov(*[combined_within_group_affinities, across_group_affinities]))
                    #print(f)
                    #print(p)
                    F_mat[i,j]=f
                    F_mat[j,i]=f
                    p_mat[i,j]=p
                    p_mat[j,i]=p
                if len(first_indices)<3 and len(second_indices)<3:
                    ## this means that these groups are both missing from this ome, so they can be considered equivalent
                    F_mat[i,j]=0
                    F_mat[j,i]=0
                    p_mat[i,j]=1
                    p_mat[j,i]=1
    ## correct p values
    #print(p_mat)
    uncorrected_p = p_mat.flatten()
    corrected_p = np.array(correct_pvalues_for_multiple_testing(uncorrected_p)).reshape(p_mat.shape)
    #print(corrected_p)
    corrected_p = np.nan_to_num(corrected_p)
    ## had to clip for visualization; led to errors in dendrogram creation 
    corrected_p = np.clip(corrected_p, 1e-310, 1)
    #print(corrected_p)
    neg_log_ten_p = -np.log10(corrected_p)
    sns.clustermap(F_mat)
    plt.savefig(out_dir+ome.name+"_affinity_F_statistic_matrix.png",
            dpi=600,
            bbox_inches='tight')
    sns.clustermap(corrected_p)
    plt.savefig(out_dir+ome.name+"_affinity_BH_p_val_matrix.png",
            dpi=600,
            bbox_inches='tight')
    sns.clustermap(neg_log_ten_p)
    plt.savefig(out_dir+ome.name+"_affinity_neg_log10_BH_p_val_matrix.png",
            dpi=600,
            bbox_inches='tight')
    print(corrected_p.shape)
    print(corrected_p)
    return(corrected_p)


def convert_sample_k_list_to_out_table(subjects, temp_sample_k_lists):
    ## create
    consensus_out_list = [[temp_sample] for temp_sample in subjects]
    for i in range(0,len(temp_sample_k_lists)):
        temp_group = temp_sample_k_lists[i]
        print(temp_group)
        for j in range(len(temp_group)):
            print(temp_group[j])
            consensus_out_list[temp_group[j]].append(i)
    return(consensus_out_list)


def write_one_hot(consensus_out_list, num_consensus, out_file):
    """
    Takes in the consensus group sample annotations:
    sample_1    0
    sample_2    1
    sample_3    0
    ...

    and writes to the out_file the one-hot encoded version. Note that this has a header, whereas the original input does not
    sample      consensus_group_0   consensus_group_1   ....
    sample_1    1                   0                   ....
    sample_2    0                   1                   ....
    sample_3    1                   0                   ....
    ....
    """
    
    ## set up the outfile header
    out_summary =[["samples"]]
    for i in range(num_consensus):
        out_summary[0].append("consensus_group_"+str(i))
    ## go through the 
    for i in range(len(consensus_out_list)):
        temp_one_hot_out_mat = np.zeros(num_consensus, dtype = int).tolist()
        temp_one_hot_out_mat[consensus_out_list[i][1]] = 1
        temp_one_hot_out_mat = [consensus_out_list[i][0]] + temp_one_hot_out_mat
        out_summary.append(temp_one_hot_out_mat)
    write_table(out_summary, out_file)
    return


def get_ome_consensus_types(ome_group_probs, temp_sample_k_lists, subjects, out_dir, one_hot = False):
    out_dir = process_dir(out_dir)
    print(ome_group_probs.shape)
    ome_type_k_lists, new_mergers, partition_dict = do_louvain_merger(ome_group_probs, temp_sample_k_lists, cutoff = 0.05)
    print('\tWe found',len(new_mergers), 'consensus groups:')
    out_summary = []
    for i in range(len(new_mergers)):
        temp_list = new_mergers[i]
        out_summary.append(["consensus_group_"+str(i), '||'.join(list(map(str, temp_list)))])
        print(out_summary[-1])
    ## get the updated lists
    consensus_out_list = convert_sample_k_list_to_out_table(subjects, ome_type_k_lists)
    write_table(deepcopy(consensus_out_list), out_dir+'sample_level_consensus_type_annotations.tsv')
    write_table(deepcopy(out_summary), out_dir+'final_cluster_level_consensus_type_annotations.tsv')
    if one_hot:
        write_one_hot(deepcopy(consensus_out_list), len(new_mergers), out_dir+'sample_level_consensus_type_annotations_one_hot.tsv')
    return(ome_type_k_lists, consensus_out_list, new_mergers)


def get_shared_list(consensus_list_1, consensus_list_2):
    ## gets the list of shared subjects
    set_1 = set(np.array(consensus_list_1)[:,0].tolist())
    set_2 = set(np.array(consensus_list_2)[:,0].tolist())
    shared = list(set_1.intersection(set_2))
    print('\tfound',len(shared),'shared subjects out of',len(set_1.union(set_2)),'total')
    return(shared)


def set_up_cont_mat(consensus_list_1, consensus_list_2):
    ## set up the empty cont mat
    ## first figure out how many groups there are in the first and second
    len_1 = len(set(np.array(consensus_list_1)[:,1].tolist()))
    len_2 = len(set(np.array(consensus_list_2)[:,1].tolist()))
    print(len_1,len_2)
    return(np.zeros((len_1,len_2)))


def subject_list_to_dict(subj_list):
    """
    Takes in 2D list of pairs of keys and values. Returns dict of them
    """
    out_dict = {}
    for line in subj_list:
        out_dict[line[0]]=line[1]
    return(out_dict)


def get_cont_mat_from_consensus_lists(consensus_list_1, consensus_list_2):
    ## first get the list of shared subjects
    common_subjects = get_shared_list(consensus_list_1, consensus_list_2)
    if len(common_subjects) < 5:
        print("\n\n\t\tWARNING! We found less than 5 shared subjects across these omes. Skipping comparison.")
        return(None)
    ## set up the empty contingency matrix
    cont_mat = set_up_cont_mat(consensus_list_1, consensus_list_2)
    subj_1_dict = subject_list_to_dict(consensus_list_1)
    subj_2_dict = subject_list_to_dict(consensus_list_2)
    for subject in common_subjects:
        row_idx = subj_1_dict[subject]
        col_idx = subj_2_dict[subject]
        cont_mat[row_idx, col_idx] += 1
    return(cont_mat)


def do_comparison(first_ome_name, second_ome_name, consensus_type_dict):
    print("\n\ncomparing",first_ome_name, "to", second_ome_name)
    consensus_list_1 = consensus_type_dict[first_ome_name]["consensus_out_list"]
    consensus_list_2 = consensus_type_dict[second_ome_name]["consensus_out_list"]
    cont_mat = get_cont_mat_from_consensus_lists(consensus_list_1, consensus_list_2)
    if not type(cont_mat)== np.ndarray:
        return
    ## make the subject-wise dictionary
    print(cont_mat)
    chi, p = chisquare(cont_mat, axis = None)
    
    ## now get cell-wise p-values
    #cell_wise_p = 
    return cont_mat, chi, p


########################################################################
########## these stats functions were borrowed from PyMINEr ############
## PyMINEr available at: www.sciencescott.com/pyminer
########################################################################

## statistics functions 
def residuals(observed, expected):
     return (observed - expected) / np.sqrt(expected)

def stdres(observed, expected):
     n = observed.sum()
     rsum, csum = margins(observed)
     v = csum * rsum * (n - rsum) * (n - csum) / n**3
     return (observed - expected) / np.sqrt(v)

def adj_res(observed,expected=None):
    ## https://journals.sagepub.com/doi/pdf/10.1177/0013164403251280
    n=observed.sum()
    rsum, csum =  margins(observed)
    #print("margins\n",rsum,csum)
    if type(expected)!=type(None):
        F=expected
    else:
        F=rsum*csum/n
    #print("F\n",F)
    res = (observed-F)/np.sqrt(F)
    z_adj_res = res/np.sqrt((1-rsum/n)*(1-csum/n))
    return(z_adj_res)

def res_to_p(residual, dof):
    chi = residual**2
    p = 1 - chi2.cdf(chi, dof)
    return(p)

def cont_to_p_independent(observed,leave_out = None):
    if leave_out != None:
        row_vect = np.arange(np.shape(observed)[0]).tolist()
        ## remove the left out index
        row_vect.pop(leave_out)
        row_vect = np.array(row_vect)
        observed = observed[row_vect,:]
    try:
        chi2, p, dof, expected = chi2_contingency(observed)
    except:
        print("\n\nWARNING: stats will be approximate due to missing data\n\n")
        temp_observed = np.array(deepcopy(observed))+1
        temp_observed = temp_observed.tolist()
        chi2, p, dof, expected = chi2_contingency(temp_observed)
    #print("traditional expected\n",expected)
    adjusted_residuals = adj_res(observed)
    p = res_to_p(adjusted_residuals,dof)
    p = correct_pvalues_for_multiple_testing(p)
    return(p, adjusted_residuals)
########################################################################

def format_res_p_merged(res_mat,p_mat):
    p_mat = p_mat.tolist()
    out_mat = deepcopy(res_mat.tolist())
    for i in range(len(out_mat)):
        for j in range(len(out_mat[i])):
            ## first change the 
            out_mat[i][j] = "{:.3g}".format(out_mat[i][j])
            p_mat[i][j] = "{:.3g}".format(p_mat[i][j])
            out_mat[i][j] = out_mat[i][j] + "\n(" +p_mat[i][j] + ")"
    return(np.array(out_mat))

def write_consensus_chi_table_results(first_ome_name, second_ome_name, cont_mat, adj_res_mat, p_mat, temp_out_dir):
    ## write the output table of statistics for the chi-square results
    out_table = [["group_1_ID", "group_2_ID", "number_in_both","adjusted_residuals","BH_corrected_p_vals","P(Grp2|Grp1)","P(Grp1|Grp2)","P(Grp2|Grp1)-P(Grp1|Grp2)"]]
    ## rows are the first group, columns are the second group
    for i in range(len(cont_mat)):## first group
        first_con_grp_name = first_ome_name+"||consensus_group_"+str(i)
        num_grp_1 = np.sum(np.array(cont_mat)[i,:])
        for j in range(len(cont_mat[i])):
            second_con_grp_name = second_ome_name+"||consensus_group_"+str(j)
            ## now append all of the results!
            num_grp_2 = np.sum(np.array(cont_mat)[:,j])
            prob_given_grp1 = cont_mat[i,j]/num_grp_1
            prob_given_grp2 = cont_mat[i,j]/num_grp_2
            prob_dif = prob_given_grp1 - prob_given_grp2
            out_table.append([first_con_grp_name, second_con_grp_name, cont_mat[i,j], adj_res_mat[i,j],p_mat[i,j], prob_given_grp1, prob_given_grp2, prob_dif])
            out_table.append([second_con_grp_name, first_con_grp_name, cont_mat[i,j], adj_res_mat[i,j],p_mat[i,j], prob_given_grp2, prob_given_grp1, -1*prob_dif])
    ## write a copy of the table
    write_table(deepcopy(out_table),os.path.join(temp_out_dir, "cross_ome_concordance_discordance_statistics.tsv"))
    ## return the contents
    return(out_table[1:])



def do_consensus_chi_analysis(first_ome_name, second_ome_name, consensus_type_dict, temp_out_dir):
    ## first make the directory for the output
    temp_out_dir = process_dir(os.path.join(temp_out_dir, first_ome_name+"_vs_"+second_ome_name))
    ## do the analysis
    cont_mat, chi, p = do_comparison(first_ome_name, second_ome_name, consensus_type_dict)
    p_mat, adjusted_residuals = cont_to_p_independent(cont_mat)
    ## plot the results
    print(chi, p)
    plt.rcParams.update({'font.size': 22})
    ax = sns.heatmap(cont_mat, annot = True)
    ax.set_title("chi="+str("{:.4g}".format(chi))+'\np='+str("{:.4g}".format(p)))
    ax.set(xlabel=second_ome_name+"\nconsensus groups", ylabel=first_ome_name+"\nconsensus groups")
    plt.savefig(temp_out_dir+"contingency_table.png",
            dpi=600,
            bbox_inches='tight')
    plt.clf()

    ax = sns.heatmap(adjusted_residuals, annot = format_res_p_merged(adjusted_residuals,p_mat), fmt = '')
    ax.set_title("chi adjusted residuals\n(BH-corrected p-value)")
    ax.set(xlabel=second_ome_name+"\nconsensus groups", ylabel=first_ome_name+"\nconsensus groups")
    plt.savefig(temp_out_dir+"chi_adjusted_residuals.png",
            dpi=600,
            bbox_inches='tight')
    plt.clf()
    print(p_mat)
    ax = sns.heatmap(p_mat, annot = True)
    ax.set_title("cell-wise adjusted p-values")
    ax.set(xlabel=second_ome_name+"\nconsensus groups", ylabel=first_ome_name+"\nconsensus groups")
    plt.savefig(temp_out_dir+"adjusted_p_values.png",
            dpi=600,
            bbox_inches='tight')
    plt.clf()

    out_table = write_consensus_chi_table_results(first_ome_name, second_ome_name, cont_mat, adjusted_residuals, p_mat, temp_out_dir)
    return(out_table)


def compare_consensus_omes(consensus_type_dict, cluster_ome_names, test_ome_names, out_dir):
    """
    This function does tests of independence for each ome.
    """
    if len(cluster_ome_names) + len(test_ome_names) <= 1:
        return
    print("checking to see which consensus types are related to each other\n")
    print("this is basically just doing Chi-squareds test of independence \non the counts for how many subjects are in each pair of consensus omes.")
    
    ## make the output directory
    consensus_comp_dir = process_dir(os.path.join(out_dir , "between_ome_consensus_group_comparisons"))

    ## go through the list of all omes & do the tests of independence
    all_cross_ome_results = [["group_1_ID", "group_2_ID", "number_in_both","adjusted_residuals","BH_corrected_p_vals","P(Grp2|Grp1)","P(Grp1|Grp2)","P(Grp2|Grp1)-P(Grp1|Grp2)"]]
    if len(cluster_ome_names) > 1:
        temp_out_dir = process_dir(os.path.join(consensus_comp_dir, "between_clustering_omes"))
        between_clustering_omes_results_table = [["group_1_ID", "group_2_ID", "number_in_both","adjusted_residuals","BH_corrected_p_vals", "P(Grp2|Grp1)","P(Grp1|Grp2)","P(Grp2|Grp1)-P(Grp1|Grp2)"]]
        ## first make the output directory
        for i in range(len(cluster_ome_names)):
            first_ome_name = cluster_ome_names[i]
            for j in range(i,len(cluster_ome_names)):
                if i != j:
                    second_ome_name = cluster_ome_names[j]
                    between_clustering_omes_results_table += do_consensus_chi_analysis(first_ome_name, second_ome_name, consensus_type_dict, temp_out_dir)
        write_table(deepcopy(between_clustering_omes_results_table),os.path.join(temp_out_dir, "cross_ome_concordance_discordance_statistics.tsv"))
        all_cross_ome_results += between_clustering_omes_results_table[1:]
    ## if there are test omes, do those comparisons too
    if len(test_ome_names) > 1:
        temp_out_dir = process_dir(os.path.join(consensus_comp_dir, "between_test_omes"))
        cross_test_ome_results = [["group_1_ID", "group_2_ID", "number_in_both","adjusted_residuals","BH_corrected_p_vals", "P(Grp2|Grp1)","P(Grp1|Grp2)","P(Grp2|Grp1)-P(Grp1|Grp2)"]]
        ## first make the output directory
        for i in range(len(test_ome_names)):
            first_ome_name = test_ome_names[i]
            for j in range(i,len(test_ome_names)):
                if i != j:
                    second_ome_name = test_ome_names[i]
                    cross_test_ome_results += do_consensus_chi_analysis(first_ome_name, second_ome_name, consensus_type_dict, temp_out_dir)
        write_table(deepcopy(cross_test_ome_results),os.path.join(temp_out_dir, "cross_ome_concordance_discordance_statistics.tsv"))
        all_cross_ome_results += cross_test_ome_results[1:]
    
    ## compare across test and clustering omes
    if len(test_ome_names) >= 1 and len(cluster_ome_names) >= 1:
        temp_out_dir = process_dir(os.path.join(consensus_comp_dir, "between_cluster_and_test_omes"))
        clust_to_test_results = [["group_1_ID", "group_2_ID", "number_in_both","adjusted_residuals","BH_corrected_p_vals", "P(Grp2|Grp1)","P(Grp1|Grp2)","P(Grp2|Grp1)-P(Grp1|Grp2)"]]
        ## first make the output directory
        for i in range(len(test_ome_names)):
            first_ome_name = test_ome_names[i]
            for j in range(len(cluster_ome_names)):
                second_ome_name = cluster_ome_names[j]
                clust_to_test_results = do_consensus_chi_analysis(first_ome_name, second_ome_name, consensus_type_dict, temp_out_dir)
        write_table(deepcopy(clust_to_test_results),os.path.join(temp_out_dir, "cross_ome_concordance_discordance_statistics.tsv"))
        all_cross_ome_results += clust_to_test_results[1:]

    ## make the output table annotating all of the subjects and their consensus groups
    write_table(deepcopy(all_cross_ome_results),os.path.join(consensus_comp_dir, "all_combined_cross_ome_concordance_discordance_statistics.tsv"))
    return


def get_cross_ome_group_mapping(source_ome_consensus_out_list, target_ome_subject_list):
    consensus_group_source_dict = {key[0]:key[1] for value, key in enumerate(source_ome_consensus_out_list)}
    out_int_list = []
    for subject in target_ome_subject_list:
        if subject in consensus_group_source_dict:
            out_int_list.append(consensus_group_source_dict[subject])
        else:
            out_int_list.append("NA")
    return(out_int_list)


def get_test_ome_differences_by_consensus_goups(out_dir, 
                                                consensus_type_dict,
                                                alpha,
                                                test_cat = [],
                                                test_num = [],
                                                processes = 1):
    """
    Take in the test omes, and assess the statistical differences of the consensus groups defined by the clustering omes
    """
    temp_top_dir = process_dir(os.path.join(out_dir, 'test_ome_differences_by_consensus_groups'))
    consensus_names = sorted(list(consensus_type_dict.keys()))
    ## go through each of the 
    for ome in test_cat + test_num:
        temp_top_test_ome_dir = process_dir(os.path.join(temp_top_dir, ome.name))
        for consensus_name in consensus_names:
            if ome.name != consensus_name:
                ## do the analysis
                temp_out_dir = process_dir(os.path.join(temp_top_test_ome_dir, consensus_name+"_consensus_groups"))
                temp_labels = consensus_type_dict[consensus_name]["consensus_out_list"]
                temp_labels = get_cross_ome_group_mapping(temp_labels, ome.subjects)
                print(temp_labels)
                ome.do_stats(temp_labels, temp_out_dir, alpha = alpha, processes = processes)
    return


def write_final_cluster_to_consensus_group_mapping_file(all_clustering_ome_consensus_list_of_list_dict, out_dir):
    """
    Takes in the dictionary that holds all of the input clustering file names & 
    """

    ## get all the unique datasets
    all_input_datasets = sorted(list(all_clustering_ome_consensus_list_of_list_dict.keys()))

    ## get all the unique final groups
    all_list_of_lists = list(all_clustering_ome_consensus_list_of_list_dict.values())
    all_final_cluster_annotations = []
    for temp_dataset_list in all_list_of_lists:
        for temp_final_cluster_list in temp_dataset_list:
            all_final_cluster_annotations += temp_final_cluster_list
    all_final_cluster_annotations = sorted(list(set(all_final_cluster_annotations)))

    ## set up the header
    header = ['final_clusters']
    for dataset in all_input_datasets:
        header.append(dataset+"_consensus_group")

    ## set up the first column
    first_col = []
    for final_cluster in all_final_cluster_annotations:
        first_col.append("sample_group_"+str(final_cluster))

    ## go through all the lists of lists & annotate for the final groups
    all_annotation_table_wo_headers = []
    for dataset in all_input_datasets:
        temp_annotations = ['NA'] * len(first_col)
        temp_consensus_list_of_lists = all_clustering_ome_consensus_list_of_list_dict[dataset]
        for i in range(len(temp_consensus_list_of_lists)):
            for j in range(len(temp_consensus_list_of_lists[i])):
                ## i is consensus group
                ## temp_consensus_list_of_lists[i][j] is final group
                temp_annotations[temp_consensus_list_of_lists[i][j]] = i
        all_annotation_table_wo_headers.append(temp_annotations)

    ## transpose the annotations
    all_annotation_table_wo_headers = np.transpose(np.array(all_annotation_table_wo_headers)).tolist()
    out_table = [header]
    for i in range(len(first_col)):
        new_line = [first_col[i]] + all_annotation_table_wo_headers[i]
        out_table.append(new_line)
    write_table(out_table, os.path.join(out_dir, "final_clusters_to_consensus_group_annotations.tsv"))
    return


def write_sample_level_final_and_consensus_group_annotations(label_pairs, censensus_out_annotations_dict, out_dir):
    """

    """
    ## get the names of all of the datasets
    ## fix label_pairs. COuldn't find where this was broken earlier
    for i in range(len(label_pairs)):
        label_pairs[i] = label_pairs[i].strip()
        label_pairs[i] = label_pairs[i].split('\t')
    ## annotate all of the datasets
    all_datasets = sorted(list(censensus_out_annotations_dict.keys()))
    consensus_group_labels = np.zeros((len(label_pairs),len(all_datasets)),dtype = '<U10')
    consensus_group_labels[:,:] = "NA"
    header = ['sample', 'final_cluster']
    for col in range(len(all_datasets)):
        dataset = all_datasets[col]
        header.append(dataset+"_consensus_group")
        temp_consensus_lists = censensus_out_annotations_dict[dataset]
        temp_lookup_dict = {value[0]:value[1] for value in temp_consensus_lists}
        for row in range(len(label_pairs)):
            temp_sample_id = label_pairs[row][0]
            if temp_sample_id in temp_lookup_dict:
                consensus_group_labels[row,col] = str(temp_lookup_dict[temp_sample_id])
    ## make the output file
    consensus_group_labels = consensus_group_labels.tolist()
    out_table = [header]
    for i in range(len(label_pairs)):
        #print(label_pairs[i])
        #print(consensus_group_labels[i])
        #print(label_pairs[i] + consensus_group_labels[i])
        out_table.append(label_pairs[i] + consensus_group_labels[i])
    write_table(out_table, os.path.join(out_dir, "sample_final_cluster_and_consensus_group_annotations.tsv"))
    return


def analyze_group_differences(meta_ome, 
                              out_dir,
                              alpha, 
                              processes = 1, 
                              test_cat = [], 
                              test_num = [],
                              one_hot = False):
    """
    Takes in the meta ome and other test omes along with a few params to do the consensus
    typing as well as the statistics within each ome and post-hoc analyses
    """
    out_dir = process_dir(out_dir)

    ## catelogue the names of the omes for organizing data later
    cluster_ome_names = [ome.name for ome in meta_ome.cat_omes_full + meta_ome.num_omes_full]
    test_ome_names = []
    if len(test_cat) + len(test_num) > 0:
        test_ome_names = [ome.name for ome in test_cat + test_num]

    ## set up the dictionary to log the consensus types
    all_consensus_out_tabels = []
    consensus_type_dict = {}

    ## do all of the analyses
    per_ome_directories = []
    all_omes = meta_ome.cat_omes_full + test_cat + meta_ome.num_omes_full + test_num
    ome_type_annotations = ["clustering"]*len(meta_ome.cat_omes_full) + ["test"] * len(test_cat) + ["clustering"] * len(meta_ome.num_omes_full) + ["test"]
    all_clustering_ome_consensus_list_of_list_dict = {}
    censensus_out_annotations_dict = {}
    for i in range(len(all_omes)):
        ome = all_omes[i]
        temp_out_dir = process_dir(out_dir + ome.name)
        per_ome_directories = []
        print("working on ome:", ome.name)
        feat_select_labels = get_per_ome_feat_select_labels(ome.subjects, meta_ome)
        ## list of lists index in first list is sample group number, entry within these are indices 
        ## of the samples within this particular ome
        temp_sample_k_lists = get_sample_k_lists(feat_select_labels)

        ## find the consensus types
        ome_group_probs = get_within_ome_similarity(temp_sample_k_lists, ome, temp_out_dir)
        consensus_group_labels, consensus_out_list, consensus_list_of_lists = get_ome_consensus_types(ome_group_probs, temp_sample_k_lists, ome.subjects, temp_out_dir, one_hot = one_hot)
        if ome_type_annotations[i] == "clustering":
            all_clustering_ome_consensus_list_of_list_dict[ome.name] = consensus_list_of_lists
            censensus_out_annotations_dict[ome.name] = consensus_out_list
        all_consensus_out_tabels.append(consensus_out_list)
        consensus_group_vect = np.array(consensus_out_list)[:,1].tolist()
        consensus_group_vect_list_int = list(map(int, deepcopy(consensus_group_vect)))

        ## log the results in the dictionary
        consensus_type_dict[ome.name] = {"ome_group_probs":ome_group_probs,
                                         "consensus_group_labels":consensus_group_labels,
                                         "consensus_out_list":consensus_out_list,
                                         "consensus_group_vect_list_int":consensus_group_vect_list_int}

        ome.do_stats(feat_select_labels, temp_out_dir, alpha = alpha, processes = processes)
        # print(feat_select_labels)
        # print(consensus_group_vect)
        ## do the same statistical analyses as before, but with the updated consensus groups
        temp_out_dir_types = process_dir(temp_out_dir + 'consensus_groups')
        ome.do_stats(consensus_group_vect_list_int, temp_out_dir_types, alpha = alpha, processes = processes, prepend_str = "consensus_group_")
        print('\n\n')

    ## TODO
    ## write the table that hosts the final group to consensus group annotations
    # use all_clustering_ome_consensus_list_of_list_dict
    write_final_cluster_to_consensus_group_mapping_file(all_clustering_ome_consensus_list_of_list_dict, out_dir)

    ## TODO
    ## write the sample_wise level annotations of final groups and consensus groups
    # use censensus_out_annotations_dict
    write_sample_level_final_and_consensus_group_annotations(meta_ome.label_pairs, censensus_out_annotations_dict, out_dir)

    compare_consensus_omes(consensus_type_dict, cluster_ome_names, test_ome_names, out_dir)
    get_test_ome_differences_by_consensus_goups(out_dir, 
                                                consensus_type_dict,
                                                alpha,
                                                test_cat = test_cat,
                                                test_num = test_num,
                                                processes = processes)
    return


##############################################################################


def parse_args():
    parser = argparse.ArgumentParser()

    ## global arguments
    parser.add_argument(
        '-cat','-c',
        default=[],
        dest='cat',
        nargs='*',
        type=str)

    parser.add_argument(
            '-missing_data','-missing','-md','-ms',
            dest='missing_str',
            help = "The string that should be interpreted as missing data. (Case sensitive)",
            default = "NA",
            type=str)

    parser.add_argument(
        '-num','-n',
        default=[],
        dest='num',
        nargs='*',
        type=str)

    parser.add_argument(
        '-test_cat','-tc',
        default=[],
        dest='test_cat',
        nargs='*',
        type=str)

    parser.add_argument(
        '-test_num','-tn',
        default=[],
        dest='test_num',
        nargs='*',
        type=str)

    parser.add_argument(
        '-processes', '-p',
        help="how many processes to use for anti-correlation based feature selection",
        default = 1,
        type=int)

    parser.add_argument(
        '-bin_size', '-bs',
        help="how many variables should be clumped together when getting the correlation matrix? Default = 5000",
        default = 5000,
        type=int)


    parser.add_argument(
        '-mem_per_node', '-mem',
        help="How much memory can each process handle (in GB)? Default = 20Gigs",
        default = 20,
        type=float)


    parser.add_argument(
        '-meta_ome_results','-ome_results','-results',
        dest='meta_ome_file',
        help = "If you already did the clustering & have the this is the 'meta_ome_results.pkl' file, and just want to annotate these results.",
        type=str)

    parser.add_argument(
        '-alpha',
        help = "The cutoff to use for signficance. Default = 0.001. With such high dimentionality as in multi-omics, you should still see plenty of results with this cutoff.",
        type=float,
        default = 0.001)

    parser.add_argument(
        '-one_hot',
        help = "This boolean flag will tell MANAclust to create a boolean 0/1 table of sample annotations for all of the consensus groups. This can be useful for feeding results into other machine learning pipelines",
        action = "store_true",
        default = False)

    parser.add_argument(
        '-out_dir',
        default = "/home/scott/bin/momcluster/lib/synthetic_data/",
        type=str)


    args = parser.parse_args()
    return(args)


def main(args):
    if args.meta_ome_file == None:
        meta_ome = cluster_omes(cat_omes = args.cat, num_omes = args.num, missing_str = args.missing_str)
    else:
        meta_ome = import_dict(args.meta_ome_file)
    test_cat_omes, test_num_omes = process_test_omes(test_cat = args.test_cat,
                                                     test_num = args.test_num, 
                                                     missing_str = args.missing_str,
                                                     processes = args.processes, 
                                                     bin_size = args.bin_size, 
                                                     process_mem = args.mem_per_node,
                                                     out_dir = args.out_dir)
    analyze_group_differences(meta_ome, 
                              args.out_dir, 
                              args.alpha, 
                              args.processes, 
                              test_cat = test_cat_omes, 
                              test_num = test_num_omes,
                              one_hot = args.one_hot)



###############################################################################

if __name__ == "__main__":
    ## set up the parser
    args = parse_args()
    main(args)




#!/usr/env python3
import os
import sys
import argparse
import numpy as np
import seaborn as sns
import pandas as pd
from sklearn.metrics import mutual_info_score
from sklearn.cluster import AffinityPropagation as ap
from sklearn.metrics import log_loss
from sklearn.preprocessing import LabelEncoder as label_maker
from sklearn.preprocessing import OneHotEncoder as onehot
from numpy import array
from numpy import argmax
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from matplotlib import pyplot as plt
from multiprocessing.dummy import Pool as ThreadPool
from sklearn.metrics import pairwise
from scipy.stats import spearmanr as spear

from mana_clust.common_functions import read_table 

euc = pairwise.euclidean_distances

##############################################################################

## hyperparameters
threads = 42


##############################################################################
def get_sd_from_beta(vect,cutoff=1):
	vect = np.array(list(vect)+list(vect*-1))
	m = np.mean(vect)
	s = np.std(vect)
	return(m+s*cutoff)


def one_hot(data):
	values = np.array(data)
	## catelogue where all of the Nones are
	none_idxs = np.where(values==None)[0]
	cat_idxs = np.where(values!=None)[0]
	integer_encoded = np.zeros((values.shape[0]))
	# integer encode
	label_encoder = LabelEncoder()
	integer_encoded = label_encoder.fit_transform(values[cat_idxs])
	
	#print(integer_encoded)
	# binary encode
	onehot_encoder = OneHotEncoder(sparse=False,categories='auto')
	integer_encoded = integer_encoded.reshape(len(integer_encoded), 1)
	onehot_encoded = onehot_encoder.fit_transform(integer_encoded)
	## make the final one hot encoded to account for Nones
	#print(onehot_encoded.shape)
	final_one_hot_encoded = np.zeros((values.shape[0],onehot_encoded.shape[1]))
	for i in range(0,cat_idxs.shape[0]):
		temp_idx = cat_idxs[i]
		final_one_hot_encoded[temp_idx] = onehot_encoded[i]
	#print(onehot_encoded)
	return(integer_encoded, final_one_hot_encoded,label_encoder)
##############################################################################
class feature():

	def __init__(self, data_vect, name):
		self.labels = list(data_vect)
		self.label_array = np.array(self.labels)
		self.label_num, self.label_embedding, self.label_encoder = one_hot(data_vect)
		self.name = name

	def get_name(self):
		return(self.name)

	def set_name(self, new_name):
		self.name = new_name

	def get_label_num(self):
		return(self.label_num)

	def get_embedding(self):
		return(self.label_embedding)

	def get_dim(self):
		return(self.get_embedding().shape)

	def get_col_dim(self):
		return(self.get_dim()[1])

	def inflate_embedding(self,col_dim):
		## check that it's a valid inflation
		if col_dim < self.get_col_dim():
			err_str = 'cannot inflate the embedding for '+self.get_name()+". Orig dim:"+str(self.get_dim())+'. Tried to set it to:'+str(col_dim)
			sys.exit(err_str)
		new_embedding = np.zeros((len(self.labels),col_dim))
		#print(new_embedding)
		for col in range(self.get_col_dim()):
			new_embedding[:,col] = self.label_embedding[:,col]
		#print(new_embedding)
		self.label_embedding = new_embedding

	def __str__(self):
		print_str = "feature object: "+self.get_name()
		print_str += "\n\t"+str(self.get_embedding())
		return(print_str)


##############################################################################
def remove_invarainat_features(in_mat):
	keep_col_vect = [True]*in_mat.shape[1]
	for i in range(1,in_mat.shape[1]):
		## after removing missing data how many unique values are there?
		feature_name = in_mat[0,i]
		temp_col_entries = in_mat[1:,i]
		temp_col_entries = temp_col_entries[temp_col_entries != None]
		num_unique_values = len(list(set(temp_col_entries)))
		if num_unique_values == 1 or num_unique_values == temp_col_entries.shape[0]:
			keep_col_vect[i]=False
			print('removing:',feature_name,num_unique_values,temp_col_entries)
	print("keeping",np.sum(keep_col_vect)-1,'features that have some variance')
	rm_vect = np.array(-1* (np.array(keep_col_vect,dtype=int)-1),dtype=bool)
	print('removing:',in_mat[0,rm_vect])
	return(in_mat[:,keep_col_vect])


def filter_none_subjects(in_mat):
	if isinstance(in_mat,list):
		in_mat = np.array(in_mat)
	## get rid of invariant features
	in_mat = remove_invarainat_features(in_mat)

	## get rid of subjects with all missing data
	
	out_mat = [in_mat[0,:].tolist()]
	all_none_cutoff = in_mat.shape[1]-1
	rm_subject = []
	for i in range(1,in_mat.shape[0]):
		temp_row = in_mat[i,1:]
		if np.sum(np.array(temp_row==None,dtype=int))==all_none_cutoff:
			rm_subject.append(in_mat[i,0])
		else:
			out_mat.append(in_mat[i,:].tolist())
	out_mat = np.array(out_mat)
	print("removed subjects:",rm_subject)
	#out_mat = remove_invarainat_features(out_mat)
	return(out_mat)


def get_randome(randome):
		## make a randomized version of this categorical ome
		## shuffle up the labels in each feature
		for i in range(1,np.shape(randome)[1]):
			np.random.shuffle(randome[1:,i])
			#randome[1:,i] = np.random.shuffle(randome[1:,i])
		print(randome)
		randome = categorical_ome(randome)
		return(randome)


class categorical_ome():
	""" this class is for any categorical ome """
	def __init__(self, in_mat, path = None, name = None):
		## rows are subjects, columns are features
		if name != None:
			self.name = name
		elif path != None:
			self.name = os.path.splitext(path)[0]
		else:
			self.name = None
		self.original_mat = filter_none_subjects(in_mat)
		self.randomized_original_mat = filter_none_subjects(in_mat)
		self.feature_names = self.original_mat[0,1:]
		self.num_features = self.feature_names.shape[0]
		self.full_cat_data = self.original_mat[1:,1:]
		self.subjects = self.original_mat[1:,0]
		self.num_subjects = np.shape(self.subjects)[0]

		self.all_features = []
		for index, temp_feature in enumerate(self.feature_names):
			#print(one_hot(full_cat_data[:,index]))
			self.all_features.append(feature(self.full_cat_data[:,index],temp_feature))

	def get_noise_factor(self,division_factor=1):
		non_zero = self.all_log_loss>1e-10
		#sns.distplot(self.all_log_loss[non_zero])
		#plt.show()
		#plt.clf()
		return(np.min(self.all_log_loss[non_zero]/division_factor))
		#return(np.mean(self.all_log_loss)/division_factor)

	def smooth_log_loss(self,smooth_iters = 1):
		noise_factor = self.get_noise_factor()
		temp_log_loss = self.all_log_loss[:]
		for i in range(smooth_iters):
			temp_log_loss+=np.random.normal(loc=0.0,scale=noise_factor,size=self.all_log_loss.shape)
		self.all_log_loss = temp_log_loss

	def process_cat_ome(self):
		## go through each feature, finding the highest col_dim
		self.all_col_dim = []
		for f in self.all_features:
			self.all_col_dim.append(f.get_col_dim())


		## go through each feature, inflating the one hot encoding matrices to enable concatenation
		self.max_col_dim = max(self.all_col_dim)
		print('max embedding dim:',self.max_col_dim)
		for f in self.all_features:
			self.all_col_dim.append(f.inflate_embedding(self.max_col_dim))

		## make the encoded matrix
		## dimtions are: [feature,subject,class within feature]
		self.all_encodings = []
		for f in self.all_features:
			#print(f.get_embedding())
			self.all_encodings.append(f.get_embedding())
		self.all_encodings = np.array(self.all_encodings)

		## get all of the log_loss distances
		self.all_log_loss = np.zeros((self.num_subjects,self.num_subjects))
		## generate the index pairs for doing the mapping
		index_pairs = []
		for i in range(0,self.num_subjects):
			for j in range(i,self.num_subjects):
				if i!=j:
					index_pairs.append([i,j])

		#################################################
		## parallelize the log losses
		print("getting log losses")

		pool = ThreadPool(threads)
		ll_pairs = pool.map(self.get_log_loss,index_pairs)
		pool.close()
		pool.join()

		#print(ll_pairs)
		#print(list(self.all_log_loss))
		print("finished getting log losses. Generating affinity matrix")
		for k in range(0,len(index_pairs)):
			i = index_pairs[k][0]
			j = index_pairs[k][1]
			self.all_log_loss[i,j] = ll_pairs[k]
			self.all_log_loss[j,i] = ll_pairs[k]

		self.smooth_log_loss()

		print(self.all_log_loss)

		# self.all_spear,dummy = spear(self.all_log_loss)
		# del dummy

		# self.affinity_matrix = -euc(self.all_spear,squared=True)
		self.affinity_matrix = -euc(self.all_log_loss, squared=True)
		self.affinity_matrix = self.affinity_matrix/(np.abs(np.min(self.affinity_matrix)/100))

	def filter_out_nones(self,temp_a,temp_b):
		""" this takes in subject a and b & filteres out any features that have missing values """
		#print('looking at a/b')
		#print(temp_a)
		#print(temp_b)
		## sum across the feature axis to figure out which features are missing a class
		temp_a_sum = np.sum(temp_a,axis=1)
		temp_b_sum = np.sum(temp_b,axis=1)
		## if either are zero, then don't consider them
		either_non = temp_a_sum * temp_b_sum
		num_features = np.sum(either_non)
		#print(temp_a_sum)
		#print(temp_b_sum)
		#print(either_non)
		#print(num_features)
		## sum them from both individuals

		mask = np.array(either_non,dtype=bool)
		#print(mask)
		#print(temp_a)
		#print(temp_b)
		temp_a = temp_a[mask,:]
		temp_b = temp_b[mask,:]
		#print(temp_a)
		#print(temp_b)
		return(temp_a, temp_b,num_features)

	def get_log_loss(self, indices):
		#global all_encodings#, all_log_loss
		i, j = indices
		#print(all_encodings[:,i,:])
		#print(all_encodings[:,j,:])
		temp_a = self.all_encodings[:,i,:]
		temp_b = self.all_encodings[:,j,:]

		temp_a, temp_b, num_features = self.filter_out_nones(temp_a,temp_b)
		ll = log_loss(temp_a,temp_b)/num_features
		#print(ll)
		#all_log_loss[i,j] = ll
		#all_log_loss[j,i] = ll
		return(ll)

	def get_good_features(self, rand_iters = 10, percent_better_than_random=0.9):
		## first calculate the observed mutual information across all features
		print('getting information for real features')
		self.real_mi_mat = self.get_cross_feature_info()
		temp_original_mat = self.randomized_original_mat[:]
		print('actual original:\n',temp_original_mat)
		#sys.exit()
		## make rand_iters number of random datasets, catelogueing the mutual
		## information across each of the features
		self.all_diff = np.zeros((rand_iters,self.num_features,self.num_features))
		for i in range(rand_iters):
			print('\tgetting rand-ome information: iter',i)
			temp_randome = get_randome(self.randomized_original_mat[:])
			temp_mi = temp_randome.get_cross_feature_info()
			self.all_diff[i,:,:] = self.real_mi_mat - temp_mi

		print('actual original:\n',self.original_mat)
		print('new original:\n',self.original_mat)
		print(temp_original_mat == self.original_mat)
		
		## now get the features that look good!
		num_better_than_random = int(percent_better_than_random * rand_iters)
		self.min_diff = np.min(self.all_diff,axis=0)
		self.max_of_min_diff = np.max(self.min_diff,axis=0)
		self.all_diff_bool = self.all_diff > 0
		self.num_times_pair_better_than_random = np.sum(self.all_diff_bool, axis=0)
		sns.clustermap(self.min_diff)
		plt.show()
		plt.clf()
		sns.distplot(self.max_of_min_diff)
		plt.show()
		plt.clf()
		self.max_of_min_diff
		self.good_feature_indices = np.where(self.max_of_min_diff>0.1)[0]#get_sd_from_beta(self.max_of_min_diff))[0]
		print('\n\nfound',self.good_feature_indices.shape[0],'good features out of',self.max_of_min_diff.shape[0])
		return(categorical_ome(self.original_mat[:,np.array([0]+list(self.good_feature_indices+1),dtype=int)]))
		# sns.clustermap(self.num_times_pair_better_than_random)
		# plt.show()
		# plt.clf()

		#per_feature_avg_signficance = np.mean(self.num_times_pair_better_than_random,percent_better_than_random)


	def get_cross_feature_info(self):
		## 
		print("\n\ngetting cross feature information:",len(self.all_features),"features")
		feature_index_pairs = []
		for i in range(0,len(self.all_features)):
			for j in range(i,len(self.all_features)):
				if i != j:
					feature_index_pairs.append([i,j])
		print('\n\n\t\t',len(feature_index_pairs),"comparisons")
		pool = ThreadPool(threads)
		mi_pairs = pool.map(self.cross_feature_mutual_info,feature_index_pairs)
		pool.close()
		pool.join()

		## log all of the mutual information across pairs of features
		pairwise_mi = np.zeros((self.num_features,self.num_features))
		for k in range(0,len(feature_index_pairs)):
			i = feature_index_pairs[k][0]
			j = feature_index_pairs[k][1]
			pairwise_mi[i,j] = mi_pairs[k]
			pairwise_mi[j,i] = mi_pairs[k]
		return(pairwise_mi)

	def cross_feature_mutual_info(self, indices):
		"""Takes in two features & make a cont table & get the mutual information"""
		feat1_idx, feat2_idx = indices
		feat1 = self.all_features[feat1_idx]
		feat2 = self.all_features[feat2_idx]

		if feat1.label_array.shape != feat2.label_array.shape:
			sys.exit("feature dimentions don't match when trying to make cont table")
		else:
			f1_not_missing = np.array(feat1.label_array!=None, dtype=int)
			f2_not_missing = np.array(feat2.label_array!=None, dtype=int)
			present_in_both = np.array(f1_not_missing * f2_not_missing, dtype=bool)
			f1_labels = feat1.label_array[present_in_both]
			f2_labels = feat2.label_array[present_in_both]
			df = pd.DataFrame({"f1":f1_labels,
				               "f2":f2_labels})
			## get the cross tab frequencies
			freq = pd.crosstab(index=df["f1"], 
			                   columns=df["f2"])

			mi = mutual_info_score(None, None, contingency = freq)
			#print(freq, mi)
		return(mi)


##############################################################################

##############################################################################
def parse_args(args):
	##########################################################################
	parser = argparse.ArgumentParser()

	## global arguments
	parser.add_argument(
		'-infile','-in','-i','-input',
		dest='infile',
		nargs='+',
		type=str)

	parser.add_argument(
			'-missing_data','-missing','-md','-ms',
			dest='missing_str',
			help = "The string that should be interpreted as missing data. (Case sensitive)",
			default = "NA",
			type=str)

	parser.add_argument(
		    '-do_clust',
		    dest='clust',
		    help='if you want to do ap clustering here directly',
		    action='store_true')

	parser.add_argument(
		    '-plot',
		    dest='plot',
		    help='if you want to make some plots',
		    action='store_true')

	args = parser.parse_args()
	return(args)
##############################################################################
## process the input file
def get_in_mat(infile,missing_str):
	in_mat = read_table(infile)
	for index, row in enumerate(in_mat):
		for idx2, element in enumerate(row):
			if row[idx2] == missing_str:
				row[idx2] = None
		in_mat[index] = row
	return(np.array(in_mat))


def do_clust(cat_ome):
	am = cat_ome.affinity_matrix
	af = ap(preference = np.min(am,axis=1),affinity="precomputed").fit(am)
	labels = af.labels_
	for idx, subject in enumerate(cat_ome.subjects):
		print(cat_ome.subjects[idx]+'\t'+str(labels[idx]))


def plot_stuff(cat_ome,do_cluster=True):
	## plot the log loss
	sns.clustermap(cat_ome.all_log_loss,row_cluster = do_cluster, col_cluster = do_cluster)
	plt.show()
	plt.clf()

	# ## get the spearman correlations
	# sns.clustermap(cat_ome.all_spear)
	# plt.show()
	# plt.clf()

	## make the normalized affinity matrix
	sns.clustermap(cat_ome.affinity_matrix, row_cluster = do_cluster, col_cluster = do_cluster)
	plt.show()
	plt.clf()

##############################################################################
## be able to do it for more than one cat ome
def do_cat_ome(infile,missing_str, do_cluster = True):
	in_mat = get_in_mat(infile, missing_str)


	## test with low none input
	cat_ome = categorical_ome(in_mat, name = infile)
	important_cat_ome = cat_ome.get_good_features()
	print("high info mat:",important_cat_ome.original_mat.shape)
	print(important_cat_ome.original_mat)
	cat_ome.process_cat_ome()
	important_cat_ome.process_cat_ome()

	if args.plot:
		print('plotting original')
		plot_stuff(cat_ome)

		print('plotting high info')
		plot_stuff(important_cat_ome)

	if args.clust:
		print("\n\n\n\ndoing clustering on full mat")
		do_clust(cat_ome)
		print("\n\n\n\ndoing clustering on important mat")
		do_clust(important_cat_ome)

	return(important_cat_ome)

def do_all_cat_omes(infile_list, missing_str):
	all_cat_omes = []
	for infile in args.infile:
		all_cat_omes.append(do_cat_ome(infile, missing_str))
	return(all_cat_omes)


##############################################################################

if __name__ == "__main__":
	## set up the parser
	args = parse_args(sys.argv)
	all_cat_omes = do_all_cat_omes(args.infile, args.missing_str)
	if args.plot:
		for cat in all_cat_omes:
			plot_stuff(cat)
	if args.clust:
		for cat in all_cat_omes:
			do_clust(cat)





import argparse
import sys
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from sklearn.cluster import AffinityPropagation as ap
import matplotlib.cm as cm
import umap
import random
import networkx as nx
import community
import pandas as pd
from sklearn.metrics import pairwise
from scipy import stats
from copy import deepcopy
from time import time

from mana_clust.common_functions import write_table, process_dir, get_sample_k_lists, save_dict
from mana_clust.mana_continuous import do_all_numeric_omes, no_p_spear, no_nan_cor, process_data_file, num_ome
from mana_clust.mana_cat import do_all_cat_omes, categorical_ome, get_in_mat
from mana_clust.mana_annotate_results import *
from mana_clust.mana_make_website import make_webpage

euc = pairwise.euclidean_distances


def get_unique_list(in_list):
    subj_list = sorted(list(set(in_list)))
    subject_idx_hash = {value:key for key, value in enumerate(subj_list)}
    return(subj_list, subject_idx_hash)


def process_ome_affinity_matrix(subject_idx_hash, all_subjects, ome):
    re_processed_am = np.zeros((len(all_subjects),len(all_subjects)))
    re_processed_am[:,:] = np.nan
    for i in range(0,len(ome.subjects)):
        subj_1_idx = subject_idx_hash[ome.subjects[i]]
        for j in range(0,len(ome.subjects)):
            subj_2_idx = subject_idx_hash[ome.subjects[j]]
            #print(subj_1_idx)
            #print(subj_2_idx)
            #print(re_processed_am)
            #print(ome.affinity_matrix)
            #print("setting:",re_processed_am[subj_1_idx,subj_2_idx],"to",ome.affinity_matrix[i,j])
            if np.isnan(ome.affinity_matrix[i,j]):
                print('found a nan')
            re_processed_am[subj_1_idx,subj_2_idx] = ome.affinity_matrix[i,j]
    return(re_processed_am)


def do_louvain_clust(am, mask_cutoff):
    ## get the mask
    mask = np.where(am<mask_cutoff)
    ## drop everything below the cutoff to -100
    am /= -mask_cutoff / 100
    ## convert the affinity matrix to weighted edges
    ## first we'll log and non-neg-ify it
    am = am * -1
    #am = np.log(am+1)## this is the log squared Euclidean distance
    ## diagonal should be zero, so we need to add one before inverting it
    am += 1
    ## second, we'll invert it
    am = 1 / am
    ## now mask it to highlight local similarity
    am[mask]=0
    ## set the diagonal back to zero
    ## whether or not to do this is debatable
    ## can't make up my mind right now
    np.fill_diagonal(am,0)
    ## normalize
    am /= np.max(am)
    print(am)
    ## make the graph
    G = nx.from_numpy_matrix(am)
    ## louvain modularity
    partition = community.best_partition(G)
    ## make the output labels
    out_labels = []
    for i in range(am.shape[0]):
        out_labels.append(partition[i])
    return(out_labels)

def do_ap_merger(am, labels):
    label_array = np.array(labels)
    ## check which groups have only one member
    singlet_group_indices = []
    for i in range(np.max(label_array)):
        if np.sum(np.array(label_array==i, dtype = int)) == 1:
            singlet_group_indices.append(i)
    ## if there are no singlet groups, we're good to go
    if len(singlet_group_indices) <= 1:
        return(labels)
    print('\nfound singlets from affinity propagation\n')
    print('\t',singlet_group_indices)
    print('\tdoing merger procedure')
    ## otherwise, regroup the singlets to make sure they belong on their own
    ## make the sample_k_lists
    temp_sample_k_lists = get_sample_k_lists(labels)
    ## then 
    final_sample_k_lists = []
    re_group_original_indices = []
    for i in range(len(temp_sample_k_lists)):
        if i in singlet_group_indices:
            re_group_original_indices += temp_sample_k_lists[i]
        else:
            final_sample_k_lists.append(temp_sample_k_lists[i])
    ## now re-cluster just the singlets
    temp_am = am[re_group_original_indices,:]
    temp_am = temp_am[:,re_group_original_indices]
    temp_labels = do_louvain_clust(temp_am, np.median(am))
    ## these are the indices from the rearranged affinity matrix subset
    ## now we need to re-transform them into their original indices
    intermediate_sample_k_lists = get_sample_k_lists(temp_labels)
    re_group_original_indices = np.array(re_group_original_indices)
    for intermediate in intermediate_sample_k_lists:
        final_sample_k_lists.append(re_group_original_indices[np.array(intermediate)].tolist())
    ## now we need to re-convert them to the linear label vectors
    final_labels = []
    for i in range(len(labels)):
        ## I know this is lazy and inefficient , but it's late
        for j in range(0,len(final_sample_k_lists)):
            if i in final_sample_k_lists[j]:
                final_labels.append(j)
    return(final_labels)


def do_clust(meta_ome, am = None, full = False):
    if type(am) == np.ndarray:
        pass
    else:
        if not full:
            am = meta_ome.affinity_matrix
        else:
            am = meta_ome.affinity_matrix_full
    print(am)
    #labels = do_louvain_clust(am, np.median(am))
    af = ap(preference = np.min(am,axis=1),affinity="precomputed").fit(am)
    labels = af.labels_
    if meta_ome == None:
        return(labels)
    labels = do_ap_merger(am, labels)
    subj_label_pairs = []
    for idx, subject in enumerate(meta_ome.all_subjects):
        subj_label_pairs.append([subject,labels[idx]])
        print(subj_label_pairs[-1][0]+'\t'+str(subj_label_pairs[-1][1]))
    ## make the label dictionary
    label_dict = {subj:group for subj, group in subj_label_pairs}
    return(labels, subj_label_pairs, label_dict)

class cluster_omes():
    def __init__(self, cat_omes = None, num_omes = None, out_dir=None, do_plot=False, missing_str = 'NA', processes = 1, bin_size = 5000, force = True, no_h5 = False, store_16bit = False, process_mem = None):
        if cat_omes == None and num_omes == None:
            print("we need to have some kind of omes fed in. Try something like cluster_omes(cat_omes = ['pat_characteristics.txt'], num_omes = ['microbiome.txt','transcriptome.txt'])")
            sys.exit()
        else:
            self.processes = processes
            self.out_dir = out_dir
            self.cat_ome_files = cat_omes
            self.num_ome_files = num_omes
            self.cat_omes_full, self.cat_omes = do_all_cat_omes(cat_omes, missing_str)
            if store_16bit:
                self.num_omes_full, self.num_omes = do_all_numeric_omes(num_omes, processes = self.processes, bin_size = bin_size, force = force, no_h5 = no_h5, dtype = np.float16, process_mem = process_mem, out_dir = out_dir)
            else:
                self.num_omes_full, self.num_omes = do_all_numeric_omes(num_omes, processes = self.processes, bin_size = bin_size, force = force, no_h5 = no_h5, process_mem = process_mem, out_dir = out_dir)
            
            ## get all of the subjects
            self.all_subjects = []
            for ome in self.cat_omes:
                self.all_subjects += ome.subjects.tolist()
            for ome in self.num_omes:
                self.all_subjects += ome.subjects.tolist()
            self.all_subjects, self.subject_idx_hash = get_unique_list(self.all_subjects)

            ############################################
            ## reprocess the FEATUER SELECTED affinity matrices
            self.affinity_matrices = []
            for ome in self.cat_omes:
                self.affinity_matrices.append(process_ome_affinity_matrix(self.subject_idx_hash, self.all_subjects, ome))
            for ome in self.num_omes:
                self.affinity_matrices.append(process_ome_affinity_matrix(self.subject_idx_hash, self.all_subjects, ome))

            ## 3D affinity matrix [ome, subject, subject]
            self.affinity_matrices = np.array(self.affinity_matrices)

            self.affinity_matrix = get_nan_norm_am(self.affinity_matrices)
            # ## get the final affinity matrix
            # self.affinity_matrix = np.nanmean(self.affinity_matrices, axis=0)

            # ## some individual pairs will be missing in both. We therefore calculate the missing value squared euclidean distance
            # ## of the combined affinity matrix
            # #self.affinity_matrix = np.nansum((self.affinity_matrix - self.affinity_matrix[:, None]) ** 2, axis=2)

            # ## Used to replace it with the mean, but the above should be better - as it yeild essentially the same results
            # self.affinity_matrix[np.isnan(self.affinity_matrix)] = np.nanmean(self.affinity_matrices)

            print("\nfinal feature selected affinity matrix:")
            print(self.affinity_matrix)

            ############################################

            ############################################
            ## reprocess the FULL affinity matrices
            self.affinity_matrices_full = []
            for ome in self.cat_omes_full:
                self.affinity_matrices_full.append(process_ome_affinity_matrix(self.subject_idx_hash, self.all_subjects, ome))
            for ome in self.num_omes_full:
                self.affinity_matrices_full.append(process_ome_affinity_matrix(self.subject_idx_hash, self.all_subjects, ome))

            ## 3D affinity matrix [ome, subject, subject]
            self.affinity_matrices_full = np.array(self.affinity_matrices_full)

            self.affinity_matrix_full = get_nan_norm_am(self.affinity_matrices_full)

            # ## get the final affinity matrix
            # self.affinity_matrix_full = np.nanmean(self.affinity_matrices_full, axis=0)

            # ## some individual pairs will be missing in both. We therefore calculate the missing value squared euclidean distance
            # ## of the combined affinity matrix
            # self.affinity_matrix_full[np.isnan(self.affinity_matrix_full)] = np.nanmean(self.affinity_matrices_full)
            # #self.affinity_matrix_full = np.nansum((self.affinity_matrix_full - self.affinity_matrix_full[:, None]) ** 2, axis=2)
            print("\nfinal full affinity matrix:")
            print(self.affinity_matrix_full)
            
            ############################################

            if do_plot:
                print(self.affinity_matrix.shape)
                sns.clustermap(self.affinity_matrix)
                plt.show()
                plt.clf()

                print(self.affinity_matrix_full.shape)
                sns.clustermap(self.affinity_matrix_full)
                plt.show()
                plt.clf()



            self.labels, self.label_pairs, self.label_dict = do_clust(self, full=False)
            self.labels_full, self.label_pairs_full, self.label_dict_full = do_clust(self, full=True)


def quantileNormalize(in_mat):
    df_input = pd.DataFrame(in_mat)
    df = df_input.copy()
    #compute rank
    dic = {}
    for col in df:
        dic.update({col : sorted(df[col])})
    sorted_df = pd.DataFrame(dic)
    rank = sorted_df.mean(axis = 1).tolist()
    #sort
    for col in df:
        t = np.searchsorted(np.sort(df[col]), df[col])
        df[col] = [rank[i] for i in t]
    return np.array(df)

def get_nan_norm_am(am, method = 'pearson'):
    ## go through each affinity matrix, and replace missing values with the median
    ## of each affinity matrix.
    for i in range(am.shape[0]):
        print("getting nan median for am matrix #",i)
        temp_med = np.nanmedian(am[i])
        ## replace the missing values with the median of this affinity matrix
        am[i][np.isnan(am[i])] = temp_med

    ## takes in the 3D affinity matrices & returns the final affinity matrix
    am = np.nanmean(am, axis=0)

    ## renormalize the matrix
    am -= np.max(am[~np.eye(am.shape[0], dtype=bool)])
    am /= -(np.min(am[~np.eye(am.shape[0], dtype=bool)])/100)

    np.fill_diagonal(am,0)


    # ## get the nan omitted correlation matrix
    # am = no_nan_cor(am, method = method)

    # ## get the -Euclidean Distance matrix
    # am = -euc(am)

    ## quantile normalize the matrix ensure that each sample has the same distribution of
    ## affinities. This helps to blunt the effect of different distributions across each
    ## given ome input

    return(am)

    # ## figure out how many acutal comparisons everything has
    # is_numeric_mat = (np.array(np.isnan(am),dtype = int)-1) * -1
    # # sns.clustermap(is_numeric_mat)
    # # plt.show()

    # ## get the average distances
    # new_am = am = -1 * np.nansum((am - am[:, None]) ** 2, axis=2)

    # ## get the number of numeric comparisons for each pair
    # num_comparisons = np.sum(is_numeric_mat * is_numeric_mat[:, None], axis = 2)
    # # sns.clustermap(num_comparisons)
    # # plt.show()
    
    # ## normalize for the number of comparisons
    # new_am /= num_comparisons

    # sns.clustermap(new_am)
    # plt.show()

    #return(new_am)

    # final_am = am[np.isnan(am)] = np.nanmean(am)
    # return(final_am)




##############################################################################
## analysis and writing results

def get_per_ome_labels(subjects, meta_ome):
    feat_labels = []
    labels_full = []
    for subj in subjects:
        feat_labels.append([subj, meta_ome.label_dict[subj]])
        labels_full.append([subj, meta_ome.label_dict_full[subj]])
    return(feat_labels, labels_full)

def write_per_ome_labels(results_dir, meta_ome):
    results_dir = process_dir(results_dir)
    for ome in meta_ome.cat_omes+meta_ome.num_omes:
        feat_labels, full_labels = get_per_ome_labels(ome.subjects, meta_ome)
        temp_out_file = results_dir + ome.name+"_labels.tsv"
        write_table(feat_labels, temp_out_file)
        temp_out_file = results_dir + ome.name+"_labels_full.tsv"
        write_table(full_labels, temp_out_file)
    return()


def prep_ome_for_saving(meta_ome):
    #meta_ome_out = deepcopy(meta_ome)
    for num_ome in meta_ome.num_omes + meta_ome.num_omes_full:
        if 'spear_out_hdf5' in dir(num_ome):
            print('removing hdf5 for saving')
            del num_ome.spear_out_hdf5
            del num_ome.spear_f
    return(meta_ome)


def write_ome_results(results_dir, meta_ome):
    results_dir = process_dir(results_dir)
    write_table(meta_ome.label_pairs, results_dir+'labels_with_feature_selection.tsv')
    write_table(meta_ome.label_pairs_full, results_dir+'labels_without_feature_selection.tsv')
    write_per_ome_labels(results_dir, meta_ome)
    save_dict(prep_ome_for_saving(meta_ome), results_dir+'meta_ome_results.pkl')
    return(results_dir)


def do_am_heatmap(out_file, ome, x_color_dict = None, y_color_dict = None):
    if x_color_dict == None:
        if isinstance(ome,np.ndarray):
            sns.clustermap(ome)
        else:
            sns.clustermap(ome.affinity_matrix)
    else:
        if y_color_dict == None:
            y_color_dict = x_color_dict
        if isinstance(ome,np.ndarray):
            sns.clustermap(ome, col_colors = x_color_dict["color_vector"],  row_colors = y_color_dict["color_vector"])
        else:
            sns.clustermap(ome.affinity_matrix, col_colors = x_color_dict["color_vector"],  row_colors = y_color_dict["color_vector"])
    plt.savefig(out_file,
                dpi=600,
                bbox_inches='tight')
    plt.clf()
    return


def plot_2d(projection, out_plot, dpi=600, color_dict = None, x_ax = '',y_ax = ''):
    global sample_id_hash, sample_k_lists, colors, args, ax, plt, temp
    print("\tPlotting",out_plot)
    sample_k_lists = color_dict["group_index_lists"]
    colors = color_dict["master_color_vect"]
    group_name_list = color_dict["group_names"]
    plt.clf()
    for i in range(0,len(sample_k_lists)):
        #temp_idxs = ids_to_idxs(sample_k_lists[i])
        temp_idxs = sample_k_lists[i]
        print(temp_idxs)
        plt.scatter(projection[temp_idxs,0],projection[temp_idxs,1],label=group_name_list[i],color=colors[i],s=10)
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.xlabel(x_ax)
    plt.ylabel(y_ax)
    if out_plot != None:
        plt.savefig(out_plot,
            dpi=dpi,
            bbox_inches='tight')
    return()


def get_umap_coords(in_mat):
    print("getting UMAP projection")
    model = umap.UMAP(n_neighbors=5,
                      min_dist=0.3,
                      metric='correlation')
    projection = model.fit_transform(in_mat)
    return(projection)


def plot_umap(out_file, in_mat = None, projection = None, color_dict = None):
    
    # if color_dict == None:
    #     ## just plot in black
    #     color_vector = ['k']*in_mat.shape[0]
    if not isinstance(projection, np.ndarray):
        projection = get_umap_coords(in_mat)

    plot_2d(projection, out_file, color_dict = color_dict, x_ax = 'UMAP1', y_ax = 'UMAP2')

    plt.savefig(out_file,
                dpi=600,
                bbox_inches='tight')
    plt.clf()
    return(projection)



def group_vector_to_color(group_vector):
    print("\nmaking group color vector:\n",group_vector[:5],"...")
    group_set_list = sorted(list(set(group_vector)))
    group_numeric_mapping = {value:key for key, value in enumerate(group_set_list)}
    group_numeric_vector = [group_numeric_mapping[group] for group in group_vector]
    colors = cm.nipy_spectral(np.arange(len(group_set_list))/len(group_set_list))
    group_color_dict = {group:colors[idx] for idx, group in enumerate(group_set_list)}
    color_vector = [group_color_dict[group] for group in group_vector]
    ## make the group_index_lists
    group_index_lists = get_sample_k_lists(group_numeric_vector)
    ## make the dictionary and return it
    full_color_dict = {"group_color_dict":group_color_dict,
                       "color_vector":color_vector,
                       "group_index_lists":group_index_lists,
                       "group_names":group_set_list,
                       "master_color_vect":colors}
    return(full_color_dict)


def plot_ome_results(meta_ome, out_dir, ground_truth_vector = None):
    ## get the colorization for the clustering results
    cluster_color_dict_feature_selected_dict = group_vector_to_color(meta_ome.labels)
    cluster_color_dict_full_dict = group_vector_to_color(meta_ome.labels_full)

    ## if there is a ground truth vector, colorize that as well
    if ground_truth_vector != None:
        ## get the color vector
        ground_truth_color_dict_full_dict = group_vector_to_color(ground_truth_vector)

    out_dir = process_dir(out_dir)
    for ome in meta_ome.cat_omes + meta_ome.num_omes:
        out_png = out_dir+ome.name+"_with_feature_selection.png"
        do_am_heatmap(out_png, ome)

    for ome in meta_ome.cat_omes_full+meta_ome.num_omes_full:
        out_png = out_dir+ome.name+"_without_feature_selection.png"
        do_am_heatmap(out_png, ome)

    do_am_heatmap(out_dir+"final_affinity_matrix_with_feature_selection.png", meta_ome.affinity_matrix, x_color_dict = cluster_color_dict_feature_selected_dict)
    do_am_heatmap(out_dir+"final_affinity_matrix_without_feature_selection.png", meta_ome.affinity_matrix_full, x_color_dict = cluster_color_dict_full_dict)
    if ground_truth_vector != None:
        do_am_heatmap(out_dir+"final_affinity_matrix_with_feature_selection_ground_truth_cols_clust_rows.png", meta_ome.affinity_matrix, x_color_dict = ground_truth_color_dict_full_dict, y_color_dict = cluster_color_dict_feature_selected_dict)
        do_am_heatmap(out_dir+"final_affinity_matrix_without_feature_selection_ground_truth_cols_clust_rows.png", meta_ome.affinity_matrix_full, x_color_dict = ground_truth_color_dict_full_dict, y_color_dict = cluster_color_dict_feature_selected_dict)


    ## do the umap projection with feature selection
    png = out_dir+"final_affinity_matrix_umap_with_feature_selection_clustering.png"
    feature_selected_umap = plot_umap(png, in_mat = meta_ome.affinity_matrix, color_dict = cluster_color_dict_feature_selected_dict)
    if ground_truth_vector != None:
        png = out_dir+"final_affinity_matrix_umap_with_feature_selection_ground_truth.png"
        plot_umap(png, projection = feature_selected_umap, color_dict = ground_truth_color_dict_full_dict)
    
    ## do the umap projection without feature selection
    png = out_dir+"final_affinity_matrix_umap_without_feature_selection_clustering.png"
    feature_selected_umap = plot_umap(png, in_mat = meta_ome.affinity_matrix_full, color_dict = cluster_color_dict_full_dict)
    if ground_truth_vector != None:
        png = out_dir+"final_affinity_matrix_umap_without_feature_selection_ground_truth.png"
        plot_umap(png, projection = feature_selected_umap, color_dict = ground_truth_color_dict_full_dict)

    return




##############################################################################

def parse_args():
    parser = argparse.ArgumentParser()

    ## global arguments
    parser.add_argument(
        '-cat','-c',
        default=[],
        dest='cat',
        nargs='*',
        type=str)

    parser.add_argument(
        '-missing_data','-missing','-md','-ms',
        dest='missing_str',
        help = "The string that should be interpreted as missing data. (Case sensitive)",
        default = "NA",
        type=str)

    parser.add_argument(
        '-num','-n',
        default=[],
        dest='num',
        nargs='*',
        type=str)

    parser.add_argument(
        '-test_cat','-tc',
        default=[],
        dest='test_cat',
        nargs='*',
        type=str)

    parser.add_argument(
        '-test_num','-tn',
        default=[],
        dest='test_num',
        nargs='*',
        type=str)

    parser.add_argument(
        '-no_post',
        dest='do_post_hoc_analyses',
        action='store_false')

    parser.add_argument(
        '-no_plots',
        dest='do_plots',
        action='store_false')

    parser.add_argument(
        '-no_force',
        dest='force',
        help="should we force recomputation of everthing or assume the temp files are correct",
        action='store_false')

    parser.add_argument(
        '-no_h5',
        help="in some scenarios it may be faster (on a cluster), to not store some of the data permanently, to minimize cross-node data transfer when going in parallel. Note that this option also requires parallel (processes >2)",
        action='store_true')


    parser.add_argument(
        '-store_16bit',
        help="If you need to save on memory, try storing the numeric omes as 16 bit floating points rather than 32 bit. Be warned that this might cause some errors down the line though. If not, you're good to go.",
        action='store_true')


    parser.add_argument(
        '-out_dir',
        default = "/home/scott/bin/mana_clust/lib/synthetic_data/",
        type=str)

    parser.add_argument(
        '-alpha',
        help="the cutoff for significance testing. Default = 0.001 - In my experience, this yeilds many good results & cuts down on false positives.",
        default = 0.001,
        type=float)

    parser.add_argument(
        '-processes', '-p',
        help="how many processes to use for anti-correlation based feature selection",
        default = 1,
        type=int)

    parser.add_argument(
        '-mem_per_node', '-mem',
        help="How much memory can each process handle (in GB)? Default = 20Gigs",
        default = 20,
        type=float)

    parser.add_argument(
        '-bin_size', '-bs',
        help="how many variables should be clumped together when getting the correlation matrix? Default = 5000",
        default = 5000,
        type=int)

    parser.add_argument(
        '-one_hot',
        help = "This boolean flag will tell MANAclust to create a boolean 0/1 table of sample annotations for all of the consensus groups. This can be useful for feeding results into other machine learning pipelines",
        action = "store_true",
        default = False)

    parser.add_argument(
        '-seed',
        help="random seed",
        default = 123456789,
        type=int)

    args = parser.parse_args()
    return(args)


def process_test_omes(test_cat = [], 
                      test_num = [], 
                      missing_str = 'NA',
                      processes = 1, 
                      bin_size = 5000, 
                      process_mem = 20,
                      out_dir = None):
    ## process the test omes if present
    test_cat_omes = []
    for cat in test_cat:
        print('\n\nprocessing test cat ome:\n',cat)
        test_cat_omes.append(categorical_ome(get_in_mat(cat, missing_str), path=cat))
        test_cat_omes[-1].process_cat_ome()
    test_num_omes = []
    for infile in test_num:
        print('processing test num ome:\n',infile)
        in_mat, feature_names, subjects = process_data_file(infile)
        test_num_omes.append(num_ome(in_mat, feature_names, subjects, all_features=True, infile = infile, processes = processes, bin_size = bin_size, process_mem = process_mem, out_dir = out_dir))
    return(test_cat_omes, test_num_omes)


def mana_clust(out_dir,
               cat = [],
               missing_str = 'NA',
               num = [],
               alpha = 0.001,
               processes = 1,
               bin_size = 5000,
               force = True,
               no_h5 = False,
               mem_per_node = 20,
               do_plots = True,
               do_post_hoc_analyses = True,
               test_cat = [],
               test_num = [],
               seed = 123456789):
    
    ## make create the outdir if it doesn't exist
    args.out_dir = process_dir(args.out_dir)
    
    ## set the seed for reproducibility
    random.seed(args.seed)
    np.random.seed(args.seed)

    ## process the meta ome and do the clustering
    meta_ome = cluster_omes(cat_omes = cat, 
                            num_omes = num, 
                            missing_str = missing_str, 
                            processes = processes, 
                            bin_size = bin_size, 
                            force = force, 
                            no_h5 = no_h5, 
                            process_mem = mem_per_node,
                            out_dir = out_dir)
    ## write the results
    write_ome_results(out_dir, meta_ome)
    
    ## make the plots if we need to
    if args.do_plots:
        plot_ome_results(meta_ome, args.out_dir)

    ## if we need to do the post-hoc analyses, do them
    if args.do_post_hoc_analyses:
        ## first process the test omes that weren't used for clustering
        test_cat_omes, test_num_omes = process_test_omes(test_cat = args.test_cat, 
                                                         test_num = args.test_num,
                                                         missing_str = args.missing_str,
                                                         processes = args.processes, 
                                                         bin_size = args.bin_size, 
                                                         process_mem = args.mem_per_node,
                                                         out_dir = args.out_dir)
        ## now do all of the analyses
        analyze_group_differences(meta_ome, 
                                  args.out_dir, 
                                  args.alpha, 
                                  args.processes, 
                                  test_cat = test_cat_omes, 
                                  test_num = test_num_omes,
                                  one_hot = args.one_hot)
    make_webpage(out_dir)
    return(meta_ome)


def main(args):
    start = time()
    mana_clust(out_dir = args.out_dir,
               cat = args.cat,
               missing_str = args.missing_str,
               num = args.num,
               alpha = args.alpha,
               processes = args.processes,
               bin_size = args.bin_size,
               force = args.force,
               no_h5 = args.no_h5,
               mem_per_node = args.mem_per_node,
               do_plots = args.do_plots,
               do_post_hoc_analyses = args.do_post_hoc_analyses,
               test_cat = args.test_cat,
               test_num = args.test_num,
               seed = args.seed)
    end = time()
    print('\n\nFinished all analyses!\n\nMANAclust took:',(end-start)/60,"minutes\n")



##############################################################################

if __name__ == "__main__":
    ## set up the parser
    args = parse_args()
    
    main(args)




#!/usr/env python3
import os
import sys
import h5py
import argparse
import numpy as np
import seaborn as sns
import pandas as pd
import scipy
from time import time
from math import ceil
from copy import deepcopy
from statistics import mean, stdev
from sklearn.metrics import mutual_info_score
from sklearn.cluster import AffinityPropagation as ap
from sklearn.metrics import log_loss
from sklearn.preprocessing import LabelEncoder as label_maker
from sklearn.preprocessing import OneHotEncoder as onehot
from numpy import array
from numpy import argmax
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from matplotlib import pyplot as plt
from multiprocessing.dummy import Pool as ThreadPool
import multiprocessing
from sklearn.metrics import pairwise
from scipy.stats import spearmanr as spear
from scipy.stats import pearsonr
from scipy.stats import rankdata
from scipy.stats import mstats_basic
from scipy.stats.mstats import f_oneway as aov
from scipy.stats import ttest_ind as ttest

from mana_clust.common_functions import read_table, write_table, correct_pvalues_for_multiple_testing, get_sample_k_lists, process_dir

euc = pairwise.euclidean_distances

threads = multiprocessing.cpu_count()



def _chk_asarray(a, axis):
    if axis is None:
        a = np.ravel(a)
        outaxis = 0
    else:
        a = np.asarray(a)
        outaxis = axis
    if a.ndim == 0:
        a = np.atleast_1d(a)
    return a, outaxis


def _contains_nan(a, nan_policy='propagate'):
    policies = ['propagate', 'raise', 'omit']
    if nan_policy not in policies:
        raise ValueError("nan_policy must be one of {%s}" %
                         ', '.join("'%s'" % s for s in policies))
    try:
        # Calling np.sum to avoid creating a huge array into memory
        # e.g. np.isnan(a).any()
        with np.errstate(invalid='ignore'):
            contains_nan = np.isnan(np.sum(a))
    except TypeError:
        # If the check cannot be properly performed we fallback to omiting
        # nan values and raising a warning. This can happen when attempting to
        # sum things that are not numbers (e.g. as in the function `mode`).
        contains_nan = False
        nan_policy = 'omit'
        warnings.warn("The input array could not be properly checked for nan "
                      "values. nan values will be ignored.", RuntimeWarning)
    if contains_nan and nan_policy == 'raise':
        raise ValueError("The input contains nan values")
    return contains_nan, nan_policy


def ss(a, axis=0):
    return(np.sum(a*a, axis))


def no_p_pear(x,y):
    x = np.asarray(x)
    y = np.asarray(y)
    n = len(x)
    mx = x.mean()
    my = y.mean()
    xm, ym = x-mx, y-my
    r_num = np.add.reduce(xm * ym)
    r_den = np.sqrt(ss(xm) * ss(ym))
    r = r_num / r_den
    # Presumably, if abs(r) > 1, then it is only some small artifact of floating
    # point arithmetic.
    r = max(min(r, 1.0), -1.0)
    return r



def no_nan_cor(a, axis = 0, method="pearson"):
    """ takes in the matrix and returns the nan omitted pearson correlation """
    if method not in ['pearson', 'spearman']:
        print('method must equal "pearson" or "spearman"')
        sys.exit()
    if axis == 1:
        a=np.transpose(a)
    a_is_num = np.array(~np.isnan(a),dtype=int)
    out_pear_mat = np.zeros((a.shape[0],a.shape[0]))
    np.fill_diagonal(out_pear_mat,1)
    for i in range(a.shape[0]):
        for j in range(i,a.shape[0]):
            if i!=j:
                usable_indices = np.array(a_is_num[i,:] * a_is_num[j,:], dtype=bool)
                if np.sum(usable_indices) > 0:
                    if method == "pearson":
                        temp_r = no_p_pear(a[i,usable_indices], a[j,usable_indices])
                    elif method == "spearman":
                        temp_r = no_p_spear(a[i,usable_indices], a[j,usable_indices])
                    out_pear_mat[i,j] =temp_r
                    out_pear_mat[j,i] =temp_r
    return(out_pear_mat)


def no_p_spear(a, b=None, axis=0, nan_policy='propagate'):
    ## nan_policy = 'omit' doesn't even behave as expected in original distribution...
    a, axisout = _chk_asarray(a, axis)
    contains_nan, nan_policy = _contains_nan(a, nan_policy)
    if contains_nan and nan_policy == 'omit':
        a = np.ma.masked_invalid(a)
        if not isinstance(b, np.ndarray):
            b = np.ma.masked_invalid(b)
        else:
            b = a
        return mstats_basic.spearmanr(a, b, axis)
    if a.size <= 1:
        return SpearmanrResult(np.nan, np.nan)
    ar = np.apply_along_axis(rankdata, axisout, a)
    br = None
    if b is not None:
        b, axisout = _chk_asarray(b, axis)
        contains_nan, nan_policy = _contains_nan(b, nan_policy)
        if contains_nan and nan_policy == 'omit':
            b = ma.masked_invalid(b)
            return mstats_basic.spearmanr(a, b, axis)
        br = np.apply_along_axis(rankdata, axisout, b)
    # n = a.shape[axisout]
    rs = np.array(np.corrcoef(ar, br, rowvar=axisout), dtype = np.float16)
    olderr = np.seterr(divide='ignore')  # rs can have elements equal to 1
    # try:
    # clip the small negative values possibly caused by rounding
    # errors before taking the square root
    # t = rs * np.sqrt(((n - 2) / ((rs + 1.0) * (1.0 - rs))).clip(0))
    # finally:
    np.seterr(**olderr)
    print(np.shape(rs), np.min(rs))
    #print('\n\n',np.min(rs),'\n\n')
    if rs.shape == (2, 2):
        return rs[1, 0]
    else:
        return rs





##################################################################
## functions for getting the empirical false positive rate
def get_Z_cutoff(vect,z=3):
    ## this is for a half distribution, so mult by -1 to make it 'normal'
    vect = np.array(vect,dtype = np.float32)
    temp_max = np.max(vect)
    temp_min = np.min(vect)
    print('min:',temp_min)
    print('max:',temp_max)
    vect = np.nan_to_num(vect)
    temp_max = np.max(vect)
    temp_min = np.min(vect)
    print('min:',temp_min)
    print('max:',temp_max)
    if temp_min < 0 and temp_max > 0:
        sys.exit('error in get_Z_cutoff: not split distirbution')
    elif temp_max > 0:
        positive = True
    elif temp_min < 0:
        positive = False
    else:
        sys.exit('all zeros fed into get_Z_cutoff')

    vect = list(vect)
    vect += list(np.array(vect)*-1)
    vect = np.array(vect)
    vect = np.clip(vect, -1, 1)
    temp_max = np.max(vect)
    temp_min = np.min(vect)
    print('min:',temp_min)
    print('max:',temp_max)
    print("# inf:",np.sum(np.array(np.isinf(vect),dtype=int)))
    print("# -inf:",np.sum(np.array(np.isneginf(vect),dtype=int)))
    vect = vect[~np.isinf(vect)]
    vect = vect[~np.isneginf(vect)]
    print("# inf:",np.sum(np.array(np.isinf(vect),dtype=int)))
    print("# -inf:",np.sum(np.array(np.isneginf(vect),dtype=int)))
    ## get the mean, & SD
    v_mean = np.mean(vect)
    v_std = np.std(vect)
    cutoff = min([0.98,v_std * z])
    if not positive:
        cutoff *= -1
    print('mean:',v_mean)
    print("StDev:",v_std)
    print('cutoff:',cutoff)
    expected_freq_of_fp = scipy.stats.norm.sf(z)
    return(cutoff, expected_freq_of_fp)


def get_shuffled(num_ome, bin_size = 5000):
    in_mat = num_ome.in_mat
    if in_mat.shape[0] <= bin_size:
        shuff_mat = deepcopy(in_mat)
    else:
        ## get a random sample of the matrix
        neg_control_sample_idxs = np.arange(in_mat.shape[0])
        np.random.shuffle(neg_control_sample_idxs)
        shuff_mat = deepcopy(in_mat[neg_control_sample_idxs[:bin_size],:])
        print(shuff_mat)
    ## now shuffle it
    for i in range(shuff_mat.shape[0]):
        np.random.shuffle(shuff_mat[i,:])
    return(shuff_mat)



############################################################################


class num_ome():
    def __init__(self, in_mat, feature_names, sample_names, 
                 name=None, 
                 all_features=False,
                 infile = None, 
                 bin_size = 5000,
                 multiplier = None,
                 processes = 1,
                 force = True, 
                 no_h5 = False,
                 process_mem = None,
                 out_dir = None):
        if name != None:
            self.name = name
        elif infile != None:
            self.name = os.path.splitext(infile)[0].split('/')[-1]
        else:
            self.name = None
        if multiplier == None:
            ## default is natural log of the number of samples
            ## clipped between 2 and 10
            log = np.log(len(sample_names)) / np.log(3.5)
            self.multiplier = min([10,max([2,log])])
            print("multiplier:",self.multiplier)
        else:
            self.multiplier = multiplier
        self.out_dir = out_dir
        self.process_mem = process_mem
        self.no_h5 = no_h5
        self.force = force
        self.processes = processes
        self.in_mat = np.nan_to_num(in_mat)
        self.feature_names = feature_names
        self.sample_names = sample_names
        self.subjects = sample_names
        self.infile = infile
        self.bin_size = bin_size
        self.all_features = all_features
        self.all_feature_names = deepcopy(self.all_features)
        ## get the bins
        self.get_good_features_and_norm(all_features = self.all_features)
        self.get_affinity_matrix()


    def plot_all_and_good_features(self):
        temp_out_dir = process_dir(os.path.join(self.out_dir,self.name))
        num_row = min([self.in_mat.shape[0], 20000])
        if num_row == self.in_mat.shape[0]:
            sns.clustermap(self.in_mat[:,:])
            plt.savefig(temp_out_dir+self.name+'_sample_of_all_features.png',
                        dpi=600,
                        bbox_inches='tight')
        else:
            sample_row = np.arange(num_row)
            np.random.shuffle(sample_row)
            sample_row = sample_row[:num_row]
            sample_row = np.array(sorted(sample_row.tolist()),dtype = int)
            try:
                sns.clustermap(self.in_mat[sample_row,:])
            except:
                print('something went wrong with plotting all features')
            else:
                plt.savefig(temp_out_dir+self.name+'_sample_of_all_features.png',
                            dpi=600,
                            bbox_inches='tight')
        plt.clf()
        try:
            sns.clustermap(self.in_mat[self.anti_cor_indices,:])
        except:
            print('something went wrong with plotting the selected features')
        else:
            plt.savefig(temp_out_dir+self.name+'_selected_features.png',
                    dpi=600,
                    bbox_inches='tight')
        plt.clf()
        return


    def get_affinity_matrix(self, out_dir = None, do_plot=True):
        ## check if there are any significant features
        if self.anti_cor_indices.shape[0] == 0:
            print("\n no selected indices!")
            self.affinity_matrix = np.zeros((self.sample_names.shape[0],self.sample_names.shape[0]))
        else:
            ## first get the spearman correlations of all samples with each other
            #print('anti-correlated indices:',self.anti_cor_indices)
            #print('subset matrix:\n',self.in_mat[self.anti_cor_indices,:])
            if do_plot and not self.all_features:
                self.plot_all_and_good_features()
            ## if there are more than ten indices, then we can do the spearman correlation first
            ## if not we have to jump right to the -Euc. Dist. Mat
            if self.anti_cor_indices.shape[0] > 10:
                sample_spear = no_p_spear(self.in_mat[self.anti_cor_indices,:])
                self.affinity_matrix = -euc(sample_spear)
            else:
                self.affinity_matrix = -euc(np.transpose(self.in_mat[self.anti_cor_indices,:]))
            #print('spear:\n',sample_spear)
            # self.affinity_matrix = -1*np.reshape(rankdata(self.affinity_matrix.flatten()),self.affinity_matrix.shape)
            # print(self.affinity_matrix)
            # print(self.affinity_matrix.shape)
            non_diag_indices = ~np.eye(self.affinity_matrix.shape[0],dtype=bool)
            non_diag_max = np.max(self.affinity_matrix[non_diag_indices])
            ## subtracting the negative, so it brings the max up to zero
            self.affinity_matrix[non_diag_indices] -= non_diag_max
            self.affinity_matrix = self.affinity_matrix/(np.min(self.affinity_matrix)/-100)
            print('affinity matrix:\n',self.affinity_matrix)
            print(np.min(self.affinity_matrix))


    def get_good_features_and_norm(self, all_features = False):
        """ takes the input matrix (features in rows, samples in columns) """
        if not all_features:
            variable_indices = self.get_variant_indices()
            self.in_mat = self.in_mat[variable_indices,:]
            self.feature_names = self.feature_names[variable_indices]
        self.bins = self.get_bins(bin_size = self.bin_size)
        
        ## min-max normalize the rows
        self.in_mat -= np.transpose(np.array([np.min(self.in_mat, axis=1)]))
        self.in_mat /= np.transpose(np.array([np.max(self.in_mat, axis=1)]))

        ## set the matrix to only the anti-correlated features
        if not all_features:
            self.get_anti_correlated_features()
            #self.in_mat = self.in_mat[anti_cor_indices,:]
        else:
            self.anti_cor_indices = np.arange(self.in_mat.shape[0])
        return()


    def is_variant_feature(self,i):
        variable = np.sum(self.in_mat[i,:] == self.in_mat[i,0]) < self.in_mat.shape[1]
        return(variable)


    def get_variant_indices(self):
        is_variable = list(map(self.is_variant_feature,list(range((self.in_mat.shape[0])))))
        return(is_variable)


    def get_bins(self, bin_size = 5000):
        print('\n\ngetting bins\n\n')
        total_vars = self.in_mat.shape[0]
        if total_vars/2 <= self.bin_size:
            ## we have to do this because there is a special case where the spearman function returns a different sized matrix if the # bins less than total vars
            print(total_vars/2,self.bin_size)
            self.bin_size = int(total_vars/2)+1
            print('new bin size:',self.bin_size)
        bins = []
        cur_bin = 0
        while cur_bin<total_vars:
            bins.append(min(cur_bin, total_vars))
            cur_bin += self.bin_size
            print(bins[-1])
        bins.append(total_vars)
        return(bins)


    def set_spear_bins(self, r, i, j):
        ## top left
        self.spear_out_hdf5[self.bins[i]:self.bins[i+1],self.bins[i]:self.bins[i+1]] = r[:self.bin_size,:self.bin_size]
        ## top right
        self.spear_out_hdf5[self.bins[i]:self.bins[i+1],self.bins[j]:self.bins[j+1]] = r[:self.bin_size,self.bin_size:]
        ## bottom left
        self.spear_out_hdf5[self.bins[j]:self.bins[j+1],self.bins[i]:self.bins[i+1],] = r[self.bin_size:,:self.bin_size]
        ## bottom right
        self.spear_out_hdf5[self.bins[j]:self.bins[j+1],self.bins[j]:self.bins[j+1]] = r[self.bin_size:,self.bin_size:]
        #print(self.spear_out_hdf5 == 0)
        #print(np.sum(self.spear_out_hdf5, axis=1))
        return()


    def add_to_spear_mat(self):
        if self.processes <= 2:
            for i in range(0,(len(self.bins)-1)):
                for j in range(i,(len(self.bins)-1)):
                    if (i!=j) or (len(self.bins) == 2):
                        print('working on',self.bins[i],self.bins[i+1],'vs',self.bins[j],self.bins[j+1])
                        r=no_p_spear(self.in_mat[self.bins[i]:self.bins[i+1],:],self.in_mat[self.bins[j]:self.bins[j+1],:], axis = 1)
                        self.set_spear_bins(r, i, j)
        else:
            ## use Ray package to do multi-processing
            import ray
            if self.process_mem != None:
                ray.init(object_store_memory=int(self.process_mem*1000000000))
            else:
                ray.init()
            @ray.remote
            def get_remote_spear(in_mat1, in_mat2, force = True):
                return(no_p_spear(in_mat1,in_mat2, axis = 1))


            ## set up the intermediate numpy matrix saving directory
            #save_dir = process_dir(self.infile+"_rho_pieces/")
            ## make the i, j pairs
            r_jobs = []
            all_i_j = []
            cur_i_j = []
            ## this will make a list leading to the paths of all of the saved matrices
            r_paths = []
            num_blocks = len(self.bins)
            total_num_block_pair_estimate = (num_blocks**2 / 2 - num_blocks)
            for i in range(0,(len(self.bins)-1)):
                for j in range(i,(len(self.bins)-1)):
                    if (i!=j) or (len(self.bins) == 2):
                        all_i_j.append([i,j])
                        #save_path = save_dir+str(i)+"_"+str(j)+".npy"
                        ## if the number of jobs exceeds the total number of processes, start a new job queue
                        if len(r_jobs) == self.processes:
                            print("submitting jobset for next",len(r_jobs))
                            temp_r_results = ray.get(r_jobs)
                            print("\t~ %",len(all_i_j)/total_num_block_pair_estimate * 100)
                            print("\tfinished gathering results, now we're saving")
                            for w in range(len(cur_i_j)):
                                i, j = cur_i_j[w]
                                start = time()
                                self.set_spear_bins(temp_r_results[w], i, j)
                                print('\t\tsetting bin took:',(time()-start),'seconds')
                            r_jobs = []
                            cur_i_j = []
                        r_jobs.append(get_remote_spear.remote(self.in_mat[self.bins[i]:self.bins[i+1],:], self.in_mat[self.bins[j]:self.bins[j+1],:], force = self.force))
                        cur_i_j.append([i, j])
            
            ## if there are leftover jobs not done yet, do them
            if len(r_jobs) > 0:
                print("submitting jobset for next",len(r_jobs))
                temp_r_results = ray.get(r_jobs)
                print("\t~ %",len(all_i_j)/total_num_block_pair_estimate * 100)
                print("\tfinished gathering results, now we're saving")
                for w in range(len(cur_i_j)):
                    i, j = cur_i_j[w]
                    self.set_spear_bins(temp_r_results[w], i, j)
                
            ray.shutdown()
            
        return()


    def make_spearman_out(self):
        self.hdf5_spear_out_file = self.infile + "_spearman.hdf5"
        print(self.hdf5_spear_out_file)
        self.spear_f = h5py.File(self.hdf5_spear_out_file, "w")
        self.spear_out_hdf5 = self.spear_f.create_dataset("infile", (self.in_mat.shape[0], self.in_mat.shape[0]), dtype=np.float16)


    def get_anti_cor_features(self):
        self.total_neg_cor = []
        self.expected_neg_cor = []
        self.sig_neg_cor = []
        self.passing = []
        self.anti_cor_indices = []
        for i in range(self.spear_out_hdf5.shape[0]):
            temp_neg_cor = np.where(self.spear_out_hdf5[i,:]<0)[0]
            self.total_neg_cor.append(len(temp_neg_cor))
            self.expected_neg_cor.append(self.total_neg_cor[-1]*self.expected_prob)
            self.sig_neg_cor.append(len(np.where(self.spear_out_hdf5[i,:]<self.neg_cutoff)[0]))
            index_passes = self.sig_neg_cor[-1] > self.expected_neg_cor[-1]*self.multiplier
            self.passing.append(index_passes)
            if index_passes:
                self.anti_cor_indices.append(i)
        self.anti_cor_indices = np.array(self.anti_cor_indices)
        for i in range(0,100):
            print(self.total_neg_cor[i],
                  self.expected_neg_cor[i],
                  self.sig_neg_cor[i])
        print('\n\nfound',len(self.anti_cor_indices),'anti-correlated features\n\n')
        return()


    def get_anti_cor_features_no_h5(self):
        ## set up the needed variables
        self.total_neg_cor = []
        self.expected_neg_cor = []
        self.sig_neg_cor = []
        self.passing = []
        self.anti_cor_indices = []

        ## set up the ray parallel function
        import ray
        if self.process_mem != None:
            ray.init(object_store_memory=int(self.process_mem*1000000000))
        else:
            ray.init()
        @ray.remote
        def ray_anti_cor(in_mat, indices, expected_prob, multiplier, neg_cutoff):
            temp_total_neg_cor = []
            temp_expected_neg_cor = []
            temp_sig_neg_cor = []
            temp_passing = []
            temp_anti_cor_indices = []
            for idx in indices:
                ## this the pearson correlation coeff of the rank transformed data
                ## note that the rank transform was done prior to parallelization to prevent 
                ## repeatedly computing this
                temp_rhos = np.zeros((in_mat.shape[0]),dtype = np.float16)
                for i in range(in_mat.shape[0]):
                    temp_rhos[i] = no_p_pear(in_mat[idx,:],in_mat[i,:])
                #print(temp_rhos)
                ## filter for the negative correlations
                temp_neg_cor = np.where(temp_rhos<0)[0]
                temp_total_neg_cor.append(len(temp_neg_cor))
                temp_expected_neg_cor.append(temp_total_neg_cor[-1]*expected_prob)
                #print(temp_neg_cor.shape[0],)
                temp_sig_neg_cor.append(len(np.where(temp_rhos<neg_cutoff)[0]))
                index_passes = temp_sig_neg_cor[-1] > temp_expected_neg_cor[-1]*multiplier
                temp_passing.append(index_passes)
                if index_passes:
                    temp_anti_cor_indices.append(idx)
            
            #######################################

            print('\t\tin current process, we found ',len(temp_anti_cor_indices),'anti-correlated indices')
            return(temp_total_neg_cor, 
                   temp_expected_neg_cor,
                   temp_sig_neg_cor,
                   temp_passing,
                   temp_anti_cor_indices)

        ## set up the ray jobs
        print('rank transforming for spearman calc')
        temp_in_mat_ranked = np.apply_along_axis(rankdata, 1, self.in_mat)
        ## decrease size as much as possible
        max_ranked = np.max(temp_in_mat_ranked)
        if max_ranked <= 255:
            print('converting ranked matrix to uint8')
            temp_in_mat_ranked = np.array(temp_in_mat_ranked, dtype = np.uint8)
        elif max_ranked <= 65535:
            print('converting ranked matrix to uint16')
            temp_in_mat_ranked = np.array(temp_in_mat_ranked, dtype = np.uint16)
        elif max_ranked <= 4294967295:
            print('converting ranked matrix to uint32')
            temp_in_mat_ranked = np.array(temp_in_mat_ranked, dtype = np.uint32)
        else:
            print('converting ranked matrix to uint64')
            temp_in_mat_ranked = np.array(temp_in_mat_ranked, dtype = np.uint64)


        ## put the matrix in shared memory 
        ray_in_mat = ray.put(temp_in_mat_ranked)
        process_lines = self.get_process_lines(list(range(self.in_mat.shape[0])), self.processes)
        r_jobs = []
        for temp_indices in process_lines:
            #r_jobs.append(ray_anti_cor.remote(ray_in_mat, temp_indices, self.expected_prob, self.multiplier, self.neg_cutoff))
            r_jobs.append(ray_anti_cor.remote(temp_in_mat_ranked, temp_indices, self.expected_prob, self.multiplier, self.neg_cutoff))

        ## execute and collate
        print('executing parallel anti-correlation based feature selection')
        all_results = ray.get(r_jobs)
        ray.shutdown()
        

        for result in all_results:
            self.total_neg_cor += result[0]
            self.expected_neg_cor += result[1]
            self.sig_neg_cor += result[2]
            self.passing += result[3]
            self.anti_cor_indices += result[4]

        print('\n\nfound',len(self.anti_cor_indices),'anti-correlated features\n\n')

        # self.total_neg_cor = np.array()
        # self.expected_neg_cor = np.array()
        # self.sig_neg_cor = np.array()
        # self.passing = np.array()
        self.anti_cor_indices = np.array(self.anti_cor_indices)

        return


    def get_anti_cor_feature_mat(self):
        ## get the full spearman matrix
        ## first set up the hdf5 file to take in the spearman values
        if self.no_h5 and self.processes >2:
            ## first make dummy variables so there aren't any errors later
            self.hdf5_spear_out_file = None
            self.spear_f = None
            self.spear_out_hdf5 = None
            self.get_anti_cor_features_no_h5()
            return()
        else:
            if self.no_h5:
                print('-no_h5 requires >=3 processes - defaulting to h5 non-parallel')
            start = time()
            self.make_spearman_out()

            ## then populate it 
            self.add_to_spear_mat()

            ## get the total number of negative correlations
            ## and get the total number less than the cutoff of these negatives
            self.get_anti_cor_features()

            ## don't forget to close the spearman file when you're done with it
            self.spear_f.close()
        return()


    def get_anti_correlated_features(self, cutoff_multiplier = 15, min_num_features = 20):
        print('original in mat shape:',self.in_mat.shape)
        if self.in_mat.shape[0] > min_num_features:
            ## get the bootstrap shuffled null distribution
            rand_mat = get_shuffled(self, bin_size = self.bin_size)
            random_rhos = no_p_spear(rand_mat,axis=1).flatten()
            print(random_rhos)
            print(np.shape(random_rhos))
            random_pos_rhos = random_rhos[random_rhos>0]
            random_neg_rhos = random_rhos[random_rhos<0]
            print(random_neg_rhos)
            self.neg_cutoff, self.expected_prob = get_Z_cutoff(random_neg_rhos)
            #self.expected_number_exceeding = int(self.in_mat.shape[0])/self.expected_freq
            #self.cutoff_for_inclusion = expected_number_exceeding * cutoff_multiplier
            #print("features must have this many negative corrs:",self.cutoff_for_inclusion)
            print('in mat shape:',self.in_mat.shape)
            start = time()
            self.get_anti_cor_feature_mat()
            print('\n\nFinished feature selection\n\n',(time()-start)/60,"minutes\n\n")
            #sys.exit()
        else:
            print("not enough variables to do feature selection - skipping with this dataset.")
            self.anti_cor_indices = np.arange(self.in_mat.shape[0])
            print('in mat shape:',self.in_mat.shape)
        return(self.anti_cor_indices)


    def single_anova(self, idx):
        list_of_group_values=[]
        for group in range(0,len(self.sample_k_list_labels)):
            if len(self.sample_k_list_labels[group]) > 0:
                list_of_group_values.append(self.in_mat[idx,self.sample_k_list_labels[group]])
        return(list(aov(*list_of_group_values)))


    def summarize_pairwise_group_differences(self, final_output, out_dir, prepend_str = "sample_group_"):
        num_diff_mat = np.zeros((len(self.sample_k_list_labels),len(self.sample_k_list_labels)))
        group_names = []
        for i in range(len(self.sample_k_list_labels)):
            group_names.append(prepend_str+str(i))
        group_index_dict = {value:key for key, value in enumerate(group_names)}
        for line in final_output:
            grp1_idx = group_index_dict[line[1]]
            grp2_idx = group_index_dict[line[2]]
            num_diff_mat[grp1_idx, grp2_idx] += 1
            num_diff_mat[grp2_idx, grp1_idx] += 1
        sns.clustermap(num_diff_mat)
        plt.savefig(out_dir+self.name+'_number_different_across_groups.png',
                dpi=600,
                bbox_inches='tight')
        plt.clf()
        return


    def do_pairwise(self, index):
        prepend_str = self.temp_prepend
        temp_name = self.feature_names[index]
        temp_data = self.in_mat[index,:]
        temp_results = []
        for i in range(0,len(self.sample_k_list_labels)):
            if len(self.sample_k_list_labels[i])!=0:
                for j in range(i,len(self.sample_k_list_labels)):
                    if i!=j:
                        if len(self.sample_k_list_labels[j])!=0:
                            first = temp_data[self.sample_k_list_labels[i]].tolist()
                            mean_1 = np.mean(temp_data[self.sample_k_list_labels[i]])
                            second = temp_data[self.sample_k_list_labels[j]].tolist()
                            mean_2 = np.mean(temp_data[self.sample_k_list_labels[j]])
                            temp_t, temp_p = ttest(first, second, equal_var = False)
                            temp_results.append([temp_name,
                                                 prepend_str+str(i),
                                                 prepend_str+str(j),
                                                 mean_1,
                                                 mean_2,
                                                 temp_t,
                                                 temp_p])
        return(temp_results)


    def get_process_lines(self, line_vect, processes):
        ## figure out how many lines each process should get
        num_lines_per_process = max([1,ceil((len(line_vect)+1) / processes)])
        print("\twe're going to be doing ~", num_lines_per_process, "lines per process")
        process_lines = [[]]## this is a list that has which lines each process is going to work on
        for i in range(len(line_vect)):
            if len(process_lines[-1]) == num_lines_per_process:
                process_lines.append([])
            process_lines[-1].append(line_vect[i])
        # print('input processes:',processes)
        # print('using processes:',len(process_lines))
        # for process in process_lines:
        #     print(process)
        # print(line_vect)
        return(process_lines)


    def do_parallel_pairwise(self, processes, prepend_str = "sample_group_"):
        process_lines = self.get_process_lines(self.global_sig_indices, processes)
        print('initializing parallel pairwise comparisons on globally significant results')

        import ray
        if self.process_mem != None:
            ray.init(object_store_memory=int(self.process_mem*1000000000))
        else:
            ray.init()
        @ray.remote
        def ray_do_pairwise(indices, in_mat_full, feature_names, sample_k_list_labels, prepend_str = "sample_group_"):
            ## in_mat is the matrix of the subset of significant indices going to this processes
            in_mat = in_mat_full[np.array(indices),:]
            temp_results = []
            for i in range(0,len(indices)):
                index = indices[i]
                temp_name = feature_names[index]
                temp_data = in_mat[i,:]
                for i in range(0,len(sample_k_list_labels)):
                    if len(sample_k_list_labels[i])!=0:
                        for j in range(i,len(sample_k_list_labels)):
                            if i!=j:
                                if len(sample_k_list_labels[j])!=0:
                                    first = temp_data[sample_k_list_labels[i]].tolist()
                                    mean_1 = np.mean(temp_data[sample_k_list_labels[i]])
                                    second = temp_data[sample_k_list_labels[j]].tolist()
                                    mean_2 = np.mean(temp_data[sample_k_list_labels[j]])
                                    temp_t, temp_p = ttest(first, second, equal_var = False)
                                    temp_results.append([temp_name,
                                                         prepend_str+str(i),
                                                         prepend_str+str(j),
                                                         mean_1,
                                                         mean_2,
                                                         temp_t,
                                                         temp_p])
            return(temp_results)

        ## now we'll start the processes
        ray_in_mat = ray.put(self.in_mat)
        r_jobs = []
        for i in range(len(process_lines)):
            ## first get the expression subset
            print('pairwise process #',i)
            if len(process_lines[i]) > 0:
                print('\t',process_lines[i][0],'through',process_lines[i][-1])
                r_jobs.append(ray_do_pairwise.remote(indices = process_lines[i], 
                                                     in_mat_full = ray_in_mat, 
                                                     feature_names = self.feature_names,
                                                     sample_k_list_labels = self.sample_k_list_labels,
                                                     prepend_str = prepend_str))
        print('\n\ninitializing parallel pairwise comparisons\n')
        all_process_aov_results = ray.get(r_jobs)
        ray.shutdown()
        ## now we have to reorganize 
        print('finished getting the parallel pairwise comparisons: collating results')
        final_results = []
        for i in range(len(all_process_aov_results)):
            final_results += all_process_aov_results[i]
        final_results = np.array(final_results)
        return(final_results)


    def ray_aov(self, processes):
        ## use Ray package to do multi-processing
        import ray
        if self.process_mem != None:
            ray.init(object_store_memory=int(self.process_mem*1000000000))
        else:
            ray.init()
        @ray.remote
        def get_ray_aov(in_mat, sample_k_list_labels):
            temp_out_results = []
            for idx in range(in_mat.shape[0]):
                list_of_group_values=[]
                for group in range(0,len(sample_k_list_labels)):
                    if len(sample_k_list_labels[group]) > 0:
                        list_of_group_values.append(in_mat[idx,sample_k_list_labels[group]])
                temp_out_results.append(list(aov(*list_of_group_values)))
            return(temp_out_results)

        process_lines = self.get_process_lines(list(range(self.in_mat.shape[0])), processes)

        ## now we'll start the processes
        r_jobs = []
        for i in range(len(process_lines)):
            ## first get the expression subset
            if i != len(process_lines)-1:
                print('process #',i)
                print('\t',process_lines[i][0],'through',process_lines[i][-1])
                temp_mat = self.in_mat[process_lines[i][0]:process_lines[i][-1]+1,:]
                print(temp_mat.shape)
                r_jobs.append(get_ray_aov.remote(temp_mat, self.sample_k_list_labels))
            else:
                print('last process #',i)
                print('\t',process_lines[i][0],'through',self.in_mat.shape[0]-1)
                temp_mat = self.in_mat[process_lines[i][0]:,:]
                print(temp_mat.shape)
                r_jobs.append(get_ray_aov.remote(temp_mat, self.sample_k_list_labels))
        print('\n\ninitializing parallel anovas\n')
        all_process_aov_results = ray.get(r_jobs)
        ray.shutdown()
        ## now we have to reorganize 
        print('finished getting the parallel anovas: collating results')
        final_aov_results = []
        for i in range(len(all_process_aov_results)):
            print('\t\t\t','in process',i,'we have',len(all_process_aov_results[i]), 'results logged')
            for j in range(len(all_process_aov_results[i])):
                final_aov_results.append(all_process_aov_results[i][j])
                #print(final_aov_results[-1])
        print('\t\t in final aov results we have',len(final_aov_results),'results')
        return(final_aov_results)


    def do_stats(self, labels, out_dir, alpha = 0.05, processes = 1, prepend_str = "sample_group_"):
        print(self.all_features)
        print("starting to do the statistics for",self.name)
        idx_list = list(range(self.in_mat.shape[0]))
        self.sample_k_list_labels = get_sample_k_lists(labels)
        arg_list = []
        for i in idx_list:
            arg_list.append(i)

        if processes <= 2:
            ## get the global statistics
            pool = ThreadPool(threads)
            global_aov = pool.map(self.single_anova,arg_list)
            pool.close()
            pool.join()
        else:
            ## this means that we're doing ray parallelization
            global_aov = self.ray_aov(processes)


        ## correct p-values
        global_aov_array = np.array(global_aov)
        self.statistic = global_aov_array[:,0]
        self.uncorrected_p = global_aov_array[:,1]
        #print(self.uncorrected_p)
        self.corrected_p = correct_pvalues_for_multiple_testing(self.uncorrected_p)
        self.global_sig_indices = list(np.where(self.corrected_p < alpha)[0])
        self.global_sig_indices += list(np.where(self.corrected_p == 0. )[0])## for whatever reason, zeros weren't coming out...
        self.global_sig_indices = np.array(sorted(list(set(self.global_sig_indices))),dtype = int)
        print('\tfound',self.global_sig_indices.shape[0] , 'significant differences')
        #print('\t\t',self.global_sig_indices)


        ## write the results as a table
        feature_names = np.concatenate((np.array(['features']),self.feature_names))
        statistic = np.concatenate((np.array(['F_statistic']),self.statistic))
        uncorrected_p  = np.concatenate((np.array(['uncorrected_p_value']),self.uncorrected_p))
        corrected_p  = np.concatenate((np.array(['BH_corrected_p_value']),self.corrected_p))
        global_out_table = np.transpose(np.concatenate((feature_names[None,:],statistic[None,:], uncorrected_p[None,:], corrected_p[None,:]), axis = 0))
        write_table(global_out_table,out_dir+self.name+"_global_statistical_differences.tsv")

        print("\tstarting to do protected pairwise analyses")
        self.temp_prepend = prepend_str
        if processes <= 2:
            pool = ThreadPool(threads)
            pairwise_results = pool.map(self.do_pairwise,self.global_sig_indices)
            pool.close()
            pool.join()
            ## reformat and correct p-vals for pairwise results
            while len(pairwise_results)>1:
                pairwise_results[0] = pairwise_results[0]+pairwise_results[1]
                del pairwise_results[1]
            if len(pairwise_results) > 0:
                pairwise_results = np.array(pairwise_results[0])
        else:
            pairwise_results = self.do_parallel_pairwise(processes, prepend_str = prepend_str)

        if len(pairwise_results) > 0:
            temp_uncorrected_p = np.array(pairwise_results[:,-1], dtype = float).tolist()
            temp_corrected_p = correct_pvalues_for_multiple_testing(temp_uncorrected_p)
            ## filter out all of the results that weren't significant
            final_output = []
            for i in range(0,len(temp_corrected_p)):
                if temp_corrected_p[i] < alpha:
                    temp_line = pairwise_results[i,:].tolist()+[temp_corrected_p[i]]
                    final_output.append(temp_line)
        else:
            final_output = []
        
        print('\ttotal of',len(final_output),'signficant post-hocs')

        ## summarize the number of differences between each group
        self.summarize_pairwise_group_differences(final_output, out_dir, prepend_str = prepend_str)

        ## write the summary table
        header = np.array(([['feature_name','group_1','group_2','mean_1','mean_2','t_statistic','uncorrected_p','BH_corrected_p']]))
        if len(final_output)>0:
            final_output = np.concatenate((header, final_output))
        write_table(final_output, out_dir+self.name+"_all_significant_postHocs.tsv")

        return



########################################################################
def process_data_file(in_path, dtype = np.float32):
    in_mat_full = np.array(read_table(in_path))
    in_mat = np.array(in_mat_full[1:,1:],dtype=dtype)
    subjects = in_mat_full[0,1:]
    feature_names = in_mat_full[1:,0]
    return(in_mat, feature_names, subjects)


def do_all_numeric_omes(infile_list, processes = 1, bin_size = 5000, force = True, no_h5 = False, process_mem = None, out_dir = None):
    all_num_omes_full = []
    all_num_omes = []
    for infile in infile_list:
        in_mat, feature_names, subjects = process_data_file(infile)
        all_num_omes.append(num_ome(in_mat, feature_names, subjects, infile = infile, processes = processes, bin_size = bin_size, force = force, no_h5 = no_h5, process_mem = process_mem, out_dir = out_dir))
        all_num_omes_full.append(num_ome(in_mat, feature_names, subjects, all_features=True, infile = infile, processes = processes, bin_size = bin_size, process_mem = process_mem, out_dir = out_dir))
    return(all_num_omes_full, all_num_omes)


def parse_args():
    parser = argparse.ArgumentParser()

    ## global arguments
    parser.add_argument(
        '-infile','-in','-i','-input',
        dest='infile',
        nargs='+',
        type=str)

    parser.add_argument(
        '-processes','-p',
        dest='processes',
        default = 1,
        type=int)

    parser.add_argument(
        '-store_16bit',
        help="If you need to save on memory, try storing the numeric omes as 16 bit floating points rather than 32 bit. Be warned that this might cause some errors down the line though. If not, you're good to go.",
        action='store_true')

    args = parser.parse_args()
    return(args)


if __name__ == "__main__":
    
    args = parse_args()
    if args.store_16bit:
        all_numeric_omes_full, all_numeric_omes = do_all_numeric_omes(args.infile, processes = args.processes, dtype=np.float16, process_mem = args.process_mem)
    else:
        all_numeric_omes_full, all_numeric_omes = do_all_numeric_omes(args.infile, processes = args.processes, process_mem = args.process_mem)
    # for ome in all_numeric_omes:
    #     ome.plot_all_and_good_features()
    




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires = fh.read()

setuptools.setup(
     name='MANAclust',  
     version='0.6.1',
     #scripts=["mana_clust/common_functions.py","mana_clust/mom_cat.py","mana_clust/mom_log_loss.py","mana_clust/mom_clust.py","mana_clust/mom_annotate_results.py", "mana_clust/mom_continuous.py"],
     author="Scott Tyler",
     author_email="scottyler89@gmail.com",
     description="Multi Affinity Network Association",
     long_description=long_description,
     long_description_content_type="text/markdown",
     install_requires = install_requires,
     url="https://scottyler892@bitbucket.org/scottyler892/manaclust",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: GNU Affero General Public License v3",
         "Operating System :: OS Independent",
     ],
 )
